#!/usr/bin/python
#
#  Copyright (c) 2016-2019, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Shell tool for controlling OpenThread NCP instances.
"""

import os
import sys
import time
import traceback
import random
import importlib

import optparse

import binascii
import socket
import struct
import string
import textwrap

import logging
import logging.config
import logging.handlers

from cmd import Cmd

from spinel.const import SPINEL
from spinel.const import kThread
from spinel.codec import WpanApi
from spinel.codec import SpinelCodec
from spinel.stream import StreamOpen
from spinel.tun import TunInterface
import spinel.config as CONFIG
import spinel.util as util

import ipaddress

__copyright__ = "Copyright (c) 2016 The OpenThread Authors."
__version__ = "0.1.0"

MASTER_PROMPT = "spinel-cli"

import io
import spinel.ipv6 as ipv6
import spinel.common as common

DEFAULT_BAUDRATE = 115200


class IcmpV6Factory(object):

    ipv6_factory = ipv6.IPv6PacketFactory(
        ehf={
            0:
                ipv6.HopByHopFactory(
                    hop_by_hop_options_factory=ipv6.HopByHopOptionsFactory(
                        options_factories={109: ipv6.MPLOptionFactory()}))
        },
        ulpf={
            58:
                ipv6.ICMPv6Factory(
                    body_factories={129: ipv6.ICMPv6EchoBodyFactory()})
        })

    def _any_identifier(self):
        return random.getrandbits(16)

    def _seq_number(self):
        seq_number = 0

        while True:
            yield seq_number
            seq_number += 1
            seq_number if seq_number < (1 << 16) else 0

    def build_icmp_echo_request(self,
                                src,
                                dst,
                                data,
                                hop_limit=64,
                                identifier=None,
                                sequence_number=None):
        identifier = self._any_identifier(
        ) if identifier is None else identifier
        sequence_number = next(
            self._seq_number()) if sequence_number is None else sequence_number

        ping_req = ipv6.IPv6Packet(
            ipv6_header=ipv6.IPv6Header(source_address=src,
                                        destination_address=dst,
                                        hop_limit=hop_limit),
            upper_layer_protocol=ipv6.ICMPv6(
                header=ipv6.ICMPv6Header(_type=ipv6.ICMP_ECHO_REQUEST, code=0),
                body=ipv6.ICMPv6EchoBody(identifier=identifier,
                                         sequence_number=sequence_number,
                                         data=data)))

        return ping_req.to_bytes()

    def from_bytes(self, data):
        return self.ipv6_factory.parse(io.BytesIO(data), common.MessageInfo())


class SpinelCliCmd(Cmd, SpinelCodec):
    """
    A command line shell for controlling OpenThread NCP nodes
    via the Spinel protocol.
    """

    VIRTUAL_TIME = os.getenv('VIRTUAL_TIME') == '1'

    icmp_factory = IcmpV6Factory()

    def _init_virtual_time(self):
        """
        compute addresses used for virtual time.
        """
        BASE_PORT = 9000
        MAX_NODES = 34
        PORT_OFFSET = int(os.getenv("PORT_OFFSET", "0"))

        self._addr = ('127.0.0.1', BASE_PORT * 2 + MAX_NODES * PORT_OFFSET)
        self._simulator_addr = ('127.0.0.1',
                                BASE_PORT + MAX_NODES * PORT_OFFSET)

    def __init__(self, stream, nodeid, vendor_module, *_a, **kw):
        if self.VIRTUAL_TIME:
            self._init_virtual_time()
        self.nodeid = nodeid
        self.tun_if = None

        self.wpan_api = WpanApi(stream, nodeid, vendor_module=vendor_module)
        self.wpan_api.queue_register(SPINEL.HEADER_DEFAULT)
        self.wpan_api.callback_register(SPINEL.PROP_STREAM_NET,
                                        self.wpan_callback)

        Cmd.__init__(self)
        Cmd.identchars = string.ascii_letters + string.digits + '-'

        if sys.stdin.isatty():
            self.prompt = MASTER_PROMPT + " > "
        else:
            self.use_rawinput = 0
            self.prompt = ""

        SpinelCliCmd.command_names.sort()

        self.history_filename = os.path.expanduser("~/.spinel-cli-history")

        try:
            import readline
            try:
                readline.read_history_file(self.history_filename)
            except IOError:
                pass
        except ImportError:
            print("Module readline unavailable")
        else:
            import rlcompleter
            if 'libedit' in readline.__doc__:
                readline.parse_and_bind('bind ^I rl_complete')
            else:
                readline.parse_and_bind('tab: complete')

        if hasattr(stream, 'pipe'):
            self.wpan_api.queue_wait_for_prop(SPINEL.PROP_LAST_STATUS,
                                              SPINEL.HEADER_ASYNC)
        self.prop_set_value(SPINEL.PROP_IPv6_ICMP_PING_OFFLOAD, 1)
        self.prop_set_value(SPINEL.PROP_THREAD_RLOC16_DEBUG_PASSTHRU, 1)

    command_names = [
        # Shell commands
        'exit',
        'quit',
        'clear',
        'history',
        'debug',
        'debug-mem',
        'v',
        'h',
        'q',

        # OpenThread CLI commands
        'help',
        'bufferinfo',
        'channel',
        'child',
        'childmax',
        'childtimeout',
        'commissioner',
        'contextreusedelay',
        'counters',
        'diag',
        'discover',
        'eidcache',
        'extaddr',
        'extpanid',
        'ifconfig',
        'ipaddr',
        'joiner',
        'keysequence',
        'leaderdata',
        'leaderweight',
        'mac',
        'macfilter',
        'masterkey',
        'mfg',
        'mode',
        'netdata',
        'networkidtimeout',
        'networkname',
        'panid',
        'parent',
        'ping',
        'prefix',
        'releaserouterid',
        'reset',
        'rloc16',
        'route',
        'router',
        'routerselectionjitter',
        'routerupgradethreshold',
        'routerdowngradethreshold',
        'scan',
        'state',
        'thread',
        'txpower',
        'version',
        'vendor',

        # OpenThread Spinel-specific commands
        'ncp-ml64',
        'ncp-ll64',
        'ncp-tun',
        'ncp-raw',
        'ncp-filter',
    ]

    @classmethod
    def wpan_callback(cls, prop, value, tid):
        consumed = False

        if prop == SPINEL.PROP_STREAM_NET:
            consumed = True

            try:
                pkt = cls.icmp_factory.from_bytes(value)

                if CONFIG.DEBUG_LOG_PKT:
                    CONFIG.LOGGER.debug(pkt)

                timenow = int(round(time.time() * 1000)) & 0xFFFFFFFF
                timestamp = (pkt.upper_layer_protocol.body.identifier << 16 |
                             pkt.upper_layer_protocol.body.sequence_number)
                timedelta = (timenow - timestamp)
                print("\n%d bytes from %s: icmp_seq=%d hlim=%d time=%dms" %
                      (len(pkt.upper_layer_protocol.body.data),
                       pkt.ipv6_header.source_address,
                       pkt.upper_layer_protocol.body.sequence_number,
                       pkt.ipv6_header.hop_limit, timedelta))
            except RuntimeError:
                pass

        return consumed

    @classmethod
    def log(cls, text):
        """ Common log handler. """
        CONFIG.LOGGER.info(text)

    def parseline(self, line):
        cmd, arg, line = Cmd.parseline(self, line)
        if cmd:
            cmd = self.short_command_name(cmd)
            line = cmd + ' ' + arg
        return cmd, arg, line

    def completenames(self, text, *ignored):
        return [
            name + ' '
            for name in SpinelCliCmd.command_names
            if name.startswith(text) or
            self.short_command_name(name).startswith(text)
        ]

    @classmethod
    def short_command_name(cls, cmd):
        return cmd.replace('-', '')

    def postloop(self):
        try:
            import readline
            try:
                readline.write_history_file(self.history_filename)
            except IOError:
                pass
        except ImportError:
            pass

    def prop_get_value(self, prop_id):
        """ Blocking helper to return value for given propery identifier. """
        return self.wpan_api.prop_get_value(prop_id)

    def prop_set_value(self, prop_id, value, py_format='B'):
        """ Blocking helper to set value for given propery identifier. """
        return self.wpan_api.prop_set_value(prop_id, value, py_format)

    def prop_insert_value(self, prop_id, value, py_format='B'):
        """ Blocking helper to insert entry for given list property. """
        return self.wpan_api.prop_insert_value(prop_id, value, py_format)

    def prop_remove_value(self, prop_id, value, py_format='B'):
        """ Blocking helper to remove entry for given list property. """
        return self.wpan_api.prop_remove_value(prop_id, value, py_format)

    def prop_get_or_set_value(self, prop_id, line, mixed_format='B'):
        """ Helper to get or set a property value based on line arguments. """
        if line:
            value = self.prep_line(line, mixed_format)
            py_format = self.prep_format(value, mixed_format)
            value = self.prop_set_value(prop_id, value, py_format)
        else:
            value = self.prop_get_value(prop_id)
        return value

    @classmethod
    def prep_line(cls, line, mixed_format='B'):
        """ Convert a command line argument to proper binary encoding (pre-pack). """
        value = line
        if line != None:
            if mixed_format == 'U':  # For UTF8, just a pass through line unmodified
                line += '\0'
                value = line.encode('utf-8')
            elif mixed_format in (
                    'D',
                    'E'):  # Expect raw data to be hex string w/o delimeters
                value = util.hex_to_bytes(line)
            elif isinstance(line, str):
                # Most everything else is some type of integer
                value = int(line, 0)
        return value

    @classmethod
    def prep_format(cls, value, mixed_format='B'):
        """ Convert a spinel format to a python pack format. """
        py_format = mixed_format
        if value == "":
            py_format = '0s'
        elif mixed_format in ('D', 'U', 'E'):
            py_format = str(len(value)) + 's'
        return py_format

    def prop_get(self, prop_id, mixed_format='B'):
        """ Helper to get a propery and output the value with Done or Error. """
        value = self.prop_get_value(prop_id)
        if value is None:
            print("Error")
            return None

        if (mixed_format == 'D') or (mixed_format == 'E'):
            print(util.hexify_str(value, ''))
        else:
            print(str(value))
        print("Done")

        return value

    def prop_set(self, prop_id, line, mixed_format='B', output=True):
        """ Helper to set a propery and output Done or Error. """
        value = self.prep_line(line, mixed_format)
        py_format = self.prep_format(value, mixed_format)
        result = self.prop_set_value(prop_id, value, py_format)

        if not output:
            return result

        if result is None:
            print("Error")
        else:
            print("Done")

        return result

    def handle_property(self, line, prop_id, mixed_format='B', output=True):
        """ Helper to set property when line argument passed, get otherwise. """
        value = self.prop_get_or_set_value(prop_id, line, mixed_format)
        if not output:
            return value

        if value is None or value == "":
            print("Error")
            return None

        if line is None or line == "":
            # Only print value on PROP_VALUE_GET
            if mixed_format == '6':
                print(str(ipaddress.IPv6Address(value)))
            elif (mixed_format == 'D') or (mixed_format == 'E'):
                print(binascii.hexlify(value).decode('utf8'))
            elif mixed_format == 'H':
                if prop_id == SPINEL.PROP_MAC_15_4_PANID:
                    print("0x%04x" % value)
                else:
                    print("%04x" % value)
            else:
                print(str(value))

        print("Done")
        return value

    def do_help(self, line):
        if line:
            cmd, _arg, _unused = self.parseline(line)
            try:
                doc = getattr(self, 'do_' + cmd).__doc__
            except AttributeError:
                doc = None
            if doc:
                self.log("%s\n" % textwrap.dedent(doc))
            else:
                self.log("No help on %s\n" % (line))
        else:
            self.print_topics(
                "\nAvailable commands (type help <name> for more information):",
                SpinelCliCmd.command_names, 15, 80)

    def do_v(self, _line):
        """
        version
            Shows detailed version information on spinel-cli tool:
        """
        self.log(MASTER_PROMPT + " ver. " + __version__)
        self.log(__copyright__)

    @classmethod
    def do_clear(cls, _line):
        """ Clean up the display. """
        os.system('reset')

    def do_history(self, _line):
        """
        history

          Show previously executed commands.
        """

        try:
            import readline
            hist = readline.get_current_history_length()
            for idx in range(1, hist + 1):
                self.log(readline.get_history_item(idx))
        except ImportError:
            pass

    def do_h(self, line):
        """ Shortcut for history. """
        self.do_history(line)

    def do_exit(self, _line):
        """ Exit the shell. """
        self.log("exit")
        return True

    def do_quit(self, line):
        """ Exit the shell. """
        return self.do_exit(line)

    def do_q(self, line):
        """ Exit the shell. """
        return self.do_exit(line)

    def do_EOF(self, _line):
        """ End of file handler for when commands are piped into shell. """
        self.log("\n")
        return True

    def emptyline(self):
        pass

    def default(self, line):
        if line[0] == "#":
            CONFIG.LOGGER.debug(line)
        else:
            CONFIG.LOGGER.info(line + ": command not found")
            # exec(line)

    def do_debug(self, line):
        """
        Enables detail logging of bytes over the wire to the radio modem.
        Usage: debug <1=enable | 0=disable>
        """

        if line != None and line != "":
            level = int(line)
        else:
            level = 0

        CONFIG.debug_set_level(level)

    def do_debugmem(self, _line):
        """ Profile python memory usage. """
        from guppy import hpy
        heap_stats = hpy()
        print(heap_stats.heap())
        print()
        print(heap_stats.heap().byrcs)

    def do_bufferinfo(self, line):
        """
        \033[1mbufferinfo\033[0m

            Get the mesh forwarder buffer info.
        \033[2m
            > bufferinfo
            total: 128
            free: 128
            6lo send: 0 0
            6lo reas: 0 0
            ip6: 0 0
            mpl: 0 0
            mle: 0 0
            arp: 0 0
            coap: 0 0
            Done
        \033[0m
        """

        result = self.prop_get_value(SPINEL.PROP_MSG_BUFFER_COUNTERS)
        if result != None:
            print("total: %d" % result[0])
            print("free: %d" % result[1])
            print("6lo send: %d %d" % result[2:4])
            print("6lo reas: %d %d" % result[4:6])
            print("ip6: %d %d" % result[6:8])
            print("mpl: %d %d" % result[8:10])
            print("mle: %d %d" % result[10:12])
            print("arp: %d %d" % result[12:14])
            print("coap: %d %d" % result[14:16])
            print("Done")
        else:
            print("Error")

    def do_channel(self, line):
        """
        \033[1mchannel\033[0m

            Get the IEEE 802.15.4 Channel value.
        \033[2m
            > channel
            11
            Done
        \033[0m
        \033[1mchannel <channel>\033[0m

            Set the IEEE 802.15.4 Channel value.
        \033[2m
            > channel 11
            Done
        \033[0m
        """
        self.handle_property(line, SPINEL.PROP_PHY_CHAN)

    def do_child(self, line):
        """\033[1m
        child list
        \033[0m
            List attached Child IDs
        \033[2m
            > child list
            1 2 3 6 7 8
            Done
        \033[0m\033[1m
        child <id>
        \033[0m
            Print diagnostic information for an attached Thread Child.
            The id may be a Child ID or an RLOC16.
        \033[2m
            > child 1
            Child ID: 1
            Rloc: 9c01
            Ext Addr: e2b3540590b0fd87
            Mode: rsn
            Net Data: 184
            Timeout: 100
            Age: 0
            LQI: 3
            RSSI: -20
            Done
        \033[0m
        """
        child_table = self.prop_get_value(SPINEL.PROP_THREAD_CHILD_TABLE)[0]

        if line == 'list':
            result = ''
            for child_data in child_table:
                child_data = child_data[0]
                child_id = child_data[1] & 0x1FF
                result += '{} '.format(child_id)
            print(result)
            print("Done")

        else:
            try:
                child_id = int(line)
                printed = False
                for child_data in child_table:
                    child_data = child_data[0]
                    id = child_data[1] & 0x1FF

                    if id == child_id:
                        mode = ''
                        if child_data[7] & 0x08:
                            mode += 'r'
                        if child_data[7] & 0x04:
                            mode += 's'
                        if child_data[7] & 0x02:
                            mode += 'd'
                        if child_data[7] & 0x01:
                            mode += 'n'

                        print("Child ID: {}".format(id))
                        print("Rloc: {:04x}".format(child_data[1]))
                        print("Ext Addr: {}".format(
                            binascii.hexlify(child_data[0])))
                        print("Mode: {}".format(mode))
                        print("Net Data: {}".format(child_data[4]))
                        print("Timeout: {}".format(child_data[2]))
                        print("Age: {}".format(child_data[3]))
                        print("LQI: {}".format(child_data[5]))
                        print("RSSI: {}".format(child_data[6]))
                        print("Done")

                        printed = True

                if not printed:
                    print("Error")
            except ValueError:
                print("Error")

    def do_childmax(self, line):
        """\033[1m
        childmax
        \033[0m
            Get the Thread Child Count Max value.
        \033[2m
            > childmax
            10
            Done
        \033[0m\033[1m
        childmax <timeout>
        \033[0m
            Set the Thread Child Count Max value.
        \033[2m
            > childmax 5
            Done
        \033[0m
        """
        self.handle_property(line, SPINEL.PROP_THREAD_CHILD_COUNT_MAX)

    def do_childtimeout(self, line):
        """\033[1m
        childtimeout
        \033[0m
            Get the Thread Child Timeout value.
        \033[2m
            > childtimeout
            300
            Done
        \033[0m\033[1m
        childtimeout <timeout>
        \033[0m
            Set the Thread Child Timeout value.
        \033[2m
            > childtimeout 300
            Done
        \033[0m
        """
        self.handle_property(line, SPINEL.PROP_THREAD_CHILD_TIMEOUT, 'L')

    def do_commissioner(self, line):
        """
        These commands are enabled when configuring with --enable-commissioner.

        \033[1m
        commissioner start
        \033[0m
            Start the Commissioner role on this node.
        \033[2m
            > commissioner start
            Done
        \033[0m\033[1m
        commissioner stop
        \033[0m
            Stop the Commissioner role on this node.
        \033[2m
            > commissioner stop
            Done
        \033[0m\033[1m
        commissioner panid <panid> <mask> <destination>
        \033[0m
            Perform panid query.
        \033[2m
            > commissioner panid 57005 4294967295 ff33:0040:fdde:ad00:beef:0:0:1
            Conflict: dead, 00000800
            Done
        \033[0m\033[1m
        commissioner energy <mask> <count> <period> <scanDuration>
        \033[0m
            Perform energy scan.
        \033[2m
            > commissioner energy 327680 2 32 1000 fdde:ad00:beef:0:0:ff:fe00:c00
            Energy: 00050000 0 0 0 0
            Done
        \033[0m
        """
        pass

    def do_contextreusedelay(self, line):
        """
        contextreusedelay

            Get the CONTEXT_ID_REUSE_DELAY value.

            > contextreusedelay
            11
            Done

        contextreusedelay <delay>

            Set the CONTEXT_ID_REUSE_DELAY value.

            > contextreusedelay 11
            Done
        """
        self.handle_property(line, SPINEL.PROP_THREAD_CONTEXT_REUSE_DELAY, 'L')

    def do_counters(self, line):
        """
        counters

            Get the supported counter names.

            > counters
            mac
            mle
            Done

        counters <countername>

            Get the counter value.

            > counters mac
            TxTotal: 10
                TxUnicast: 3
                TxBroadcast: 7
                TxAckRequested: 3
                TxAcked: 3
                TxNoAckRequested: 7
                TxData: 10
                TxDataPoll: 0
                TxBeacon: 0
                TxBeaconRequest: 0
                TxOther: 0
                TxRetry: 0
                    TxDirectRetrySuccess: [ 0:2, 1:2, 2:1 ]
                    TxDirectMaxRetryExpiry: 1
                    TxIndirectRetrySuccess: [ 0:0 ]
                    TxIndirectMaxRetryExpiry: 1
                TxErrCca: 0
                TxAbort: 0
                TxErrBusyChannel: 0
            RxTotal: 2
                RxUnicast: 1
                RxBroadcast: 1
                RxData: 2
                RxDataPoll: 0
                RxBeacon: 0
                RxBeaconRequest: 0
                RxOther: 0
                RxAddressFiltered: 0
                RxDestAddrFiltered: 0
                RxDuplicated: 0
                RxErrNoFrame: 0
                RxErrNoUnknownNeighbor: 0
                RxErrInvalidSrcAddr: 0
                RxErrSec: 0
                RxErrFcs: 0
                RxErrOther: 0
            Done
            > counters mle
            Role Disabled: 0
            Role Detached: 1
            Role Child: 0
            Role Router: 0
            Role Leader: 1
            Attach Attempts: 1
            Partition Id Changes: 1
            Better Partition Attach Attempts: 0
            Parent Changes: 0
            Done

        counters <countername> reset

            Reset the counter value.

            > counters mac reset
            Done
            > counters mle reset
            Done

        """

        params = line.split(" ")

        if params[0] == "mac":

            if len(params) == 1:
                histogram = None
                result = self.prop_get_value(SPINEL.PROP_CNTR_ALL_MAC_COUNTERS)
                caps_list = self.prop_get_value(SPINEL.PROP_CAPS)

                for caps in caps_list[0]:
                    if SPINEL.CAP_MAC_RETRY_HISTOGRAM == caps[0][0]:
                        histogram = self.prop_get_value(
                            SPINEL.PROP_CNTR_MAC_RETRY_HISTOGRAM)

                if result != None:
                    counters_tx = result[0][0]
                    counters_rx = result[1][0]

                    print("TxTotal: %d" % counters_tx[0])
                    print("    TxUnicast: %d" % counters_tx[1])
                    print("    TxBroadcast: %d" % counters_tx[2])
                    print("    TxAckRequested: %d" % counters_tx[3])
                    print("    TxAcked: %d" % counters_tx[4])
                    print("    TxNoAckRequested: %d" % counters_tx[5])
                    print("    TxData: %d" % counters_tx[6])
                    print("    TxDataPoll: %d" % counters_tx[7])
                    print("    TxBeacon: %d" % counters_tx[8])
                    print("    TxBeaconRequest: %d" % counters_tx[9])
                    print("    TxOther: %d" % counters_tx[10])
                    print("    TxRetry: %d" % counters_tx[11])
                    if histogram != None:
                        histogram_direct = histogram[0][0]
                        if len(histogram_direct) != 0:
                            print("        TxDirectRetrySuccess: [", end='')
                            for retry in range(len(histogram_direct)):
                                print(" %d:%s" %
                                      (retry, histogram_direct[retry][0]),
                                      end=',' if retry !=
                                      (len(histogram_direct) - 1) else " ]\n")
                    print("        TxDirectMaxRetryExpiry: %s" %
                          (counters_tx[15][0]))
                    if histogram != None:
                        histogram_indirect = histogram[1][0]
                        if len(histogram_indirect) != 0:
                            print("        TxIndirectRetrySuccess: [", end='')
                            for retry in range(len(histogram_indirect)):
                                print(" %d:%s" %
                                      (retry, histogram_indirect[retry][0]),
                                      end=',' if retry !=
                                      (len(histogram_indirect) - 1) else " ]\n")
                    print("        TxIndirectMaxRetryExpiry: %s" %
                          (counters_tx[16][0]))
                    print("    TxErrCca: %d" % counters_tx[12])
                    print("    TxAbort: %d" % counters_tx[13])
                    print("    TxErrBusyChannel: %d" % counters_tx[14])
                    print("RxTotal: %d" % counters_rx[0])
                    print("    RxUnicast: %d" % counters_rx[1])
                    print("    RxBroadcast: %d" % counters_rx[2])
                    print("    RxData: %d" % counters_rx[3])
                    print("    RxDataPoll: %d" % counters_rx[4])
                    print("    RxBeacon: %d" % counters_rx[5])
                    print("    RxBeaconRequest: %d" % counters_rx[6])
                    print("    RxOther: %d" % counters_rx[7])
                    print("    RxAddressFiltered: %d" % counters_rx[8])
                    print("    RxDestAddrFiltered: %d" % counters_rx[9])
                    print("    RxDuplicated: %d" % counters_rx[10])
                    print("    RxErrNoFrame: %d" % counters_rx[11])
                    print("    RxErrNoUnknownNeighbor: %d" % counters_rx[12])
                    print("    RxErrInvalidSrcAddr: %d" % counters_rx[13])
                    print("    RxErrSec: %d" % counters_rx[14])
                    print("    RxErrFcs: %d" % counters_rx[15])
                    print("    RxErrOther: %d" % counters_rx[16])
                    print("Done")
                else:
                    print("Error")

            elif len(params) == 2:
                if params[1] == "reset":
                    self.prop_set_value(SPINEL.PROP_CNTR_ALL_MAC_COUNTERS, 1)
                    self.prop_set_value(SPINEL.PROP_CNTR_MAC_RETRY_HISTOGRAM, 1)
                    print("Done")
            else:
                print("Error")

        elif params[0] == "mle":

            if len(params) == 1:
                result = self.prop_get_value(SPINEL.PROP_CNTR_MLE_COUNTERS)
                if result != None:
                    print("Role Disabled: %d" % result[0])
                    print("Role Detached: %d" % result[1])
                    print("Role Child: %d" % result[2])
                    print("Role Router: %d" % result[3])
                    print("Role Leader: %d" % result[4])
                    print("Attach Attempts: %d" % result[5])
                    print("Partition Id Changes: %d" % result[6])
                    print("Better Partition Attach Attempts: %d" % result[7])
                    print("Parent Changes: %d" % result[8])
                    print("Done")
                else:
                    print("Error")

            elif len(params) == 2:
                if params[1] == "reset":
                    self.prop_set_value(SPINEL.PROP_CNTR_MLE_COUNTERS, 1)
                    print("Done")
            else:
                print("Error")

        elif params[0] is None or params[0] == "":
            print("mac")
            print("mle")
            print("Done")
        else:
            print("Error")

    def do_discover(self, line):
        """
        discover [channel]

             Perform an MLE Discovery operation.

        channel: The channel to discover on. If no channel is provided,
        the discovery will cover all valid channels.

        > discover
        | J | Network Name     | Extended PAN     | PAN  | MAC Address      | Ch | dBm | LQI |
        +---+------------------+------------------+------+------------------+----+-----+-----+
        | 0 | OpenThread       | dead00beef00cafe | ffff | f1d92a82c8d8fe43 | 11 | -20 |   0 |
        Done
        """
        pass

    def do_eidcache(self, line):
        """
        eidcache

            Print the EID-to-RLOC cache entries.

            > eidcache
            fdde:ad00:beef:0:bb1:ebd6:ad10:f33 ac00
            fdde:ad00:beef:0:110a:e041:8399:17cd 6000
            Done
        """
        pass

    def do_extaddr(self, line):
        """
        extaddr

            Get the IEEE 802.15.4 Extended Address.

            > extaddr
            dead00beef00cafe
            Done

        extaddr <extaddr>

            Set the IEEE 802.15.4 Extended Address.

            > extaddr dead00beef00cafe
            dead00beef00cafe
            Done
        """
        self.handle_property(line, SPINEL.PROP_MAC_15_4_LADDR, 'E')

    def do_extpanid(self, line):
        """
        extpanid

            Get the Thread Extended PAN ID value.

            > extpanid
            dead00beef00cafe
            Done

        extpanid <extpanid>

            Set the Thread Extended PAN ID value.

            > extpanid dead00beef00cafe
            Done
        """
        self.handle_property(line, SPINEL.PROP_NET_XPANID, 'D')

    def do_joiner(self, line):
        """
        These commands are enabled when configuring with --enable-joiner.

        joiner start <pskd> <provisioningUrl>

            Start the Joiner role.

            * pskd: Pre-Shared Key for the Joiner.
            * provisioningUrl: Provisioning URL for the Joiner (optional).

            This command will cause the device to perform an MLE Discovery and
            initiate the Thread Commissioning process.

            > joiner start PSK
            Done

        joiner stop

            Stop the Joiner role.

            > joiner stop
            Done
        """
        PSKd = ""

        params = line.split(" ")
        if len(params) > 0:
            sub_command = params[0]
        if len(params) > 1:
            PSKd = params[1]

        PSKd = self.prep_line(PSKd, 'U')

        if sub_command == "":
            pass

        elif sub_command == "start":
            py_format = self.prep_format(PSKd, 'U')
            self.prop_set_value(SPINEL.PROP_MESHCOP_JOINER_CREDENTIAL, PSKd,
                                py_format)
            self.prop_set_value(SPINEL.PROP_MESHCOP_JOINER_ENABLE, 1)
            print("Done")
            return

        elif sub_command == "stop":
            self.prop_set_value(SPINEL.PROP_MESHCOP_JOINER_ENABLE, 0)
            print("Done")
            return

        print("Error")

    def complete_ifconfig(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for ifconfig command. """
        map_sub_commands = ('up', 'down')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_ifconfig(self, line):
        """
        ifconfig up

            Bring up the IPv6 interface.

            > ifconfig up
            Done

        ifconfig down

            Bring down the IPv6 interface.

            > ifconfig down
            Done

        ifconfig

            Show the status of the IPv6 interface.

            > ifconfig
            down
            Done
        """

        params = line.split(" ")

        if params[0] == "":
            value = self.prop_get_value(SPINEL.PROP_NET_IF_UP)
            if value != None:
                map_arg_value = {
                    0: "down",
                    1: "up",
                }
                print(map_arg_value[value])

        elif params[0] == "up":
            self.prop_set(SPINEL.PROP_NET_IF_UP, '1')
            return

        elif params[0] == "down":
            self.prop_set(SPINEL.PROP_NET_IF_UP, '0')
            return

        print("Done")

    def complete_ipaddr(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for ipaddr command. """
        map_sub_commands = ('add', 'remove')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_ipaddr(self, line):
        """
        ipaddr

            List all IPv6 addresses assigned to the Thread interface.

            > ipaddr
            fdde:ad00:beef:0:0:ff:fe00:0
            fe80:0:0:0:0:ff:fe00:0
            fdde:ad00:beef:0:558:f56b:d688:799
            fe80:0:0:0:f3d9:2a82:c8d8:fe43
            Done

        ipaddr add <ipaddr>

            Add an IPv6 address to the Thread interface.

            > ipaddr add 2001::dead:beef:cafe
            Done

        ipaddr del <ipaddr>

            Delete an IPv6 address from the Thread interface.

            > ipaddr del 2001::dead:beef:cafe
            Done
        """
        params = line.split(" ")
        valid = 1
        preferred = 1
        flags = 0
        # always use /64, as prefix.network.prefixlen returns /128.
        prefix_len = 64

        num = len(params)
        if num > 1:
            ipaddr = params[1]
            prefix = ipaddress.IPv6Interface(str(ipaddr))
            arr = prefix.ip.packed

        if params[0] == "":
            addrs = self.wpan_api.get_ipaddrs()
            for addr in addrs:
                print(str(addr))

        elif params[0] == "add":
            arr += self.wpan_api.encode_fields('CLLC', prefix_len, valid,
                                               preferred, flags)

            self.prop_insert_value(SPINEL.PROP_IPV6_ADDRESS_TABLE, arr,
                                   str(len(arr)) + 's')

            if self.tun_if:
                self.tun_if.addr_add(ipaddr)

        elif params[0] == "remove":
            arr += self.wpan_api.encode_fields('CLLC', prefix_len, valid,
                                               preferred, flags)

            self.prop_remove_value(SPINEL.PROP_IPV6_ADDRESS_TABLE, arr,
                                   str(len(arr)) + 's')
            if self.tun_if:
                self.tun_if.addr_del(ipaddr)

        print("Done")

    def do_keysequence(self, line):
        """
        keysequence counter

            Get the Thread Key Sequence Counter.

            > keysequence counter
            10
            Done

        keysequence counter <counter>

            Set the Thread Key Sequence Counter.

            > keysequence counter 10
            Done

        keysequence guardtime

            Get the thrKeySwitchGuardTime (in hours).

            > keysequence guardtime
            0
            Done

        keysequence guardtime <guardtime>

            Set the thrKeySwitchGuardTime (in hours).

            > keysequence guardtime 0
            Done
        """

        args = line.split(" ")

        if args[0] == "counter":
            newline = line.replace("counter", "")
            self.handle_property(newline, SPINEL.PROP_NET_KEY_SEQUENCE_COUNTER,
                                 'L')

        elif args[0] == "guardtime":
            newline = line.replace("guardtime", "")
            self.handle_property(newline, SPINEL.PROP_NET_KEY_SWITCH_GUARDTIME,
                                 'L')

    def do_leaderdata(self, line):
        """
        leaderdata

            Get the Thread network Leader Data.

            > leaderdata
            Partition ID: 1987912443
            Weighting: 64
            Data Version: 4
            Stable Data Version: 129
            Leader Router ID: 47
            Done
        """
        partition_id = self.prop_get_value(SPINEL.PROP_NET_PARTITION_ID)
        weighting = self.prop_get_value(SPINEL.PROP_THREAD_LEADER_WEIGHT)
        data_version = self.prop_get_value(
            SPINEL.PROP_THREAD_NETWORK_DATA_VERSION)
        stable_version = self.prop_get_value(
            SPINEL.PROP_THREAD_STABLE_NETWORK_DATA_VERSION)
        leader_id = self.prop_get_value(SPINEL.PROP_THREAD_LEADER_RID)

        if partition_id   is None or \
           weighting      is None or \
           data_version   is None or \
           stable_version is None or \
           leader_id is None:
            print("Error")
        else:
            print("Partition ID: %d" % partition_id)
            print("Weighting: %d" % weighting)
            print("Data Version: %d" % data_version)
            print("Stable Data Version: %d" % stable_version)
            print("Leader Router ID: %d" % leader_id)
            print("Done")

    def do_leaderweight(self, line):
        """
        leaderweight

            Get the Thread Leader Weight.

            > leaderweight
            128
            Done

        leaderweight <weight>

            Set the Thread Leader Weight.

            > leaderweight 128
            Done
        """
        self.handle_property(line, SPINEL.PROP_THREAD_LOCAL_LEADER_WEIGHT)

    def do_masterkey(self, line):
        """
        masterkey

            Get the Thread Master Key value.

            > masterkey
            00112233445566778899aabbccddeeff
            Done

        masterkey <key>

            Set the Thread Master Key value.

            > masterkey 00112233445566778899aabbccddeeff
            Done
        """
        self.handle_property(line, SPINEL.PROP_NET_MASTER_KEY, 'D')

    def do_mfg(self, line):
        """
        mfg <diagnostic command>

        Check all the factory diagnostic commands here:
        https://github.com/openthread/openthread/blob/master/src/core/diags/README.md

        For example:

            Start the diagnostic module.

                > mfg start
                start diagnostics mode
                status 0x00

            Retrieved radio statistics.

                > mfg stats
                received packets: 0
                sent packets: 0
                first received packet: rssi=0, lqi=0
                last received packet: rssi=0, lqi=0
        """
        result = self.prop_set(SPINEL.PROP_NEST_STREAM_MFG, line, 'U', False)
        if result != None:
            print(result.rstrip())
        else:
            print("Error")

    def do_mode(self, line):
        """
        mode

            Get the Thread Device Mode value.

              r: rx-on-when-idle
              d: Full Function Device
              n: Full Network Data

            > mode
            rdn
            Done

        mode [rdn]

            Set the Thread Device Mode value.

              r: rx-on-when-idle
              d: Full Function Device
              n: Full Network Data

            > mode rsdn
            Done
        """
        map_arg_value = {
            0x00: "-",
            0x01: "n",
            0x02: "d",
            0x03: "dn",
            0x08: "r",
            0x09: "rn",
            0x0A: "rd",
            0x0B: "rdn",
        }

        map_arg_name = {
            "-": "0",
            "n": 0x01,
            "d": 0x02,
            "dn": 0x03,
            "r": 0x08,
            "rn": 0x09,
            "rd": 0x0A,
            "rdn": 0x0B,
        }

        if line:
            try:
                # remap string state names to integer
                line = map_arg_name[line]
            except KeyError:
                print("Error")
                return

        result = self.prop_get_or_set_value(SPINEL.PROP_THREAD_MODE, line)
        if result != None:
            if not line:
                print(map_arg_value[result])
            print("Done")
        else:
            print("Error")

    def do_netdata(self, line):
        """
        netdata

            Register local network data with Thread Leader.

            > netdata register
            Done
        """
        params = line.split(" ")
        if params[0] == "register":
            self.prop_set_value(SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE,
                                1)
            self.handle_property("0",
                                 SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE)

    def do_networkidtimeout(self, line):
        """
        networkidtimeout

            Get the NETWORK_ID_TIMEOUT parameter used in the Router role.

            > networkidtimeout
            120
            Done

        networkidtimeout <timeout>

            Set the NETWORK_ID_TIMEOUT parameter used in the Router role.

            > networkidtimeout 120
            Done
        """
        self.handle_property(line, SPINEL.PROP_THREAD_NETWORK_ID_TIMEOUT)

    def do_networkname(self, line):
        """
        networkname

            Get the Thread Network Name.

            > networkname
            OpenThread
            Done

        networkname <name>

            Set the Thread Network Name.

            > networkname OpenThread
            Done
        """
        self.handle_property(line, SPINEL.PROP_NET_NETWORK_NAME, 'U')

    def do_panid(self, line):
        """
        panid

            Get the IEEE 802.15.4 PAN ID value.

            > panid
            0xdead
            Done

        panid <panid>

            Set the IEEE 802.15.4 PAN ID value.

            > panid 0xdead
            Done
        """
        self.handle_property(line, SPINEL.PROP_MAC_15_4_PANID, 'H')

    def do_parent(self, line):
        """
        parent

            Get the addresses of the parent node.

            > parent
            Ext Addr: 3ad35f9846ceb9c7
            Rloc: bc00
            Done
        """
        ext_addr, rloc = self.prop_get_value(SPINEL.PROP_THREAD_PARENT)

        if ext_addr is None or\
           rloc is None:
            print("Error")
        else:
            print("Ext Addr: {}".format(binascii.hexlify(ext_addr)))
            print("Rloc: {:04x}".format(rloc))

    def do_ping(self, line):
        """
        ping <ipaddr> [size] [count] [interval]

            Send an ICMPv6 Echo Request.

            > ping fdde:ad00:beef:0:558:f56b:d688:799
            16 bytes from fdde:ad00:beef:0:558:f56b:d688:799: icmp_seq=1 hlim=64 time=28ms
        """
        params = line.split(" ")
        addr = "::1"
        _size = "56"
        _count = "1"
        _interval = "1"
        if len(params) > 0:
            addr = params[0]
        if len(params) > 1:
            _size = params[1]
        if len(params) > 2:
            _count = params[2]
        if len(params) > 3:
            _interval = params[3]

        try:
            # Generate local ping packet and send directly via spinel.
            ml64 = self.prop_get_value(SPINEL.PROP_IPV6_ML_ADDR)
            ml64 = str(ipaddress.IPv6Address(ml64))
            timenow = int(round(time.time() * 1000)) & 0xFFFFFFFF
            data = bytearray(int(_size))

            ping_req = self.icmp_factory.build_icmp_echo_request(
                ml64,
                addr,
                data,
                identifier=(timenow >> 16),
                sequence_number=(timenow & 0xffff))

            self.wpan_api.ip_send(ping_req)
            # Let handler print result
        except:
            print("Fail")
            print(traceback.format_exc())

    def complete_prefix(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for prefix command. """
        map_sub_commands = ('add', 'remove')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_prefix(self, line):
        """
        prefix add <prefix> [pvdcsr] [prf]

            Add a valid prefix to the Network Data.

              p: Preferred flag
              a: Stateless IPv6 Address Autoconfiguration flag
              d: DHCPv6 IPv6 Address Configuration flag
              c: DHCPv6 Other Configuration flag
              r: Default Route flag
              o: On Mesh flag
              s: Stable flag
              prf: Default router preference, which may be 'high', 'med', or 'low'.
            > prefix add 2001:dead:beef:cafe::/64 paros med
            Done

        prefix remove <prefix>

            Invalidate a prefix in the Network Data.

            > prefix remove 2001:dead:beef:cafe::/64
            Done
        """
        params = line.split(" ")
        stable = 0
        flags = 0
        arr = ""

        num = len(params)
        if num > 1:
            prefix = ipaddress.IPv6Interface(str(params[1]))
            arr = prefix.ip.packed

        if num > 2:
            map_param_to_flag = {
                'p': kThread.PrefixPreferredFlag,
                'a': kThread.PrefixSlaacFlag,
                'd': kThread.PrefixDhcpFlag,
                'c': kThread.PrefixConfigureFlag,
                'r': kThread.PrefixDefaultRouteFlag,
                'o': kThread.PrefixOnMeshFlag,
            }
            for char in params[2]:
                if char == 's':
                    stable = 1  # Stable flag
                else:
                    flag = map_param_to_flag.get(char, None)
                    if flag is not None:
                        flags |= flag

        if num > 3:
            map_arg_name = {
                "high": 2,
                "med": 1,
                "low": 0,
            }
            prf = map_arg_name[params[3]]
            flags |= (prf << kThread.PrefixPreferenceOffset)

        if params[0] == "":
            self.prop_get_value(SPINEL.PROP_THREAD_ON_MESH_NETS)

        elif params[0] == "add":
            arr += self.wpan_api.encode_fields('CbC', prefix.network.prefixlen,
                                               stable, flags)

            self.prop_set_value(SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE,
                                1)
            self.prop_insert_value(SPINEL.PROP_THREAD_ON_MESH_NETS, arr,
                                   str(len(arr)) + 's')

        elif params[0] == "remove":
            arr += self.wpan_api.encode_fields('CbC', prefix.network.prefixlen,
                                               stable, flags)

            self.prop_set_value(SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE,
                                1)
            self.prop_remove_value(SPINEL.PROP_THREAD_ON_MESH_NETS, arr,
                                   str(len(arr)) + 's')

        print("Done")

    def do_releaserouterid(self, line):
        """
        releaserouterid <routerid>

            Release a Router ID that has been allocated by the device in the Leader role.

            > releaserouterid 16
            Done
        """
        if line:
            value = int(line)
            self.prop_remove_value(SPINEL.PROP_THREAD_ACTIVE_ROUTER_IDS, value)
        print("Done")

    def do_rloc16(self, line):
        """
        rloc16

            Get the Thread RLOC16 value.

            > rloc16
            0xdead
            Done
        """
        self.handle_property(line, SPINEL.PROP_THREAD_RLOC16, 'H')

    def do_reset(self, line):
        """
        reset

            Reset the NCP.

            > reset
        """
        self.wpan_api.cmd_reset()

        self.prop_set_value(SPINEL.PROP_IPv6_ICMP_PING_OFFLOAD, 1)
        self.prop_set_value(SPINEL.PROP_THREAD_RLOC16_DEBUG_PASSTHRU, 1)

    def complete_route(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for route command. """
        map_sub_commands = ('add', 'remove')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_route(self, line):
        """
        route add <prefix> [s] [prf]

            Add a valid prefix to the Network Data.

              s: Stable flag
              prf: Default Router Preference, which may be: 'high', 'med', or 'low'.

            > route add 2001:dead:beef:cafe::/64 s med
            Done

        route remove <prefix>

            Invalidate a prefix in the Network Data.

            > route remove 2001:dead:beef:cafe::/64
            Done
        """
        params = line.split(" ")
        stable = 0
        prf = 0

        num = len(params)
        if num > 1:
            prefix = ipaddress.IPv6Interface(str(params[1]))
            arr = prefix.ip.packed

        if params[0] == "":
            self.prop_get_value(SPINEL.PROP_THREAD_LOCAL_ROUTES)

        elif params[0] == "add":
            arr += self.wpan_api.encode_fields('CbC', prefix.network.prefixlen,
                                               stable, prf)

            self.prop_set_value(SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE,
                                1)
            self.prop_insert_value(SPINEL.PROP_THREAD_LOCAL_ROUTES, arr,
                                   str(len(arr)) + 's')

        elif params[0] == "remove":
            self.prop_set_value(SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE,
                                1)
            self.prop_remove_value(SPINEL.PROP_THREAD_LOCAL_ROUTES, arr,
                                   str(len(arr)) + 's')

        print("Done")

    def do_router(self, line):
        """
        router list

            List allocated Router IDs

            > router list
            8 24 50
            Done

        router <id>

            Print diagnostic information for a Thread Router.
            The id may be a Router ID or an RLOC16.

            > router 50
            Alloc: 1
            Router ID: 50
            Rloc: c800
            Next Hop: c800
            Link: 1
            Ext Addr: e2b3540590b0fd87
            Cost: 0
            LQI In: 3
            LQI Out: 3
            Age: 3
            Done

            > router 0xc800
            Alloc: 1
            Router ID: 50
            Rloc: c800
            Next Hop: c800
            Link: 1
            Ext Addr: e2b3540590b0fd87
            Cost: 0
            LQI In: 3
            LQI Out: 3
            Age: 7
            Done
        """
        pass

    def do_routerselectionjitter(self, line):
        """
        routerselectionjitter

            Get the ROUTER_SELECTION_JITTER value.

            > routerselectionjitter
            120
            Done

        routerselectionjitter <threshold>

            Set the ROUTER_SELECTION_JITTER value.

            > routerselectionjitter 120
            Done
        """
        self.handle_property(line, SPINEL.PROP_THREAD_ROUTER_SELECTION_JITTER)

    def do_routerupgradethreshold(self, line):
        """
        routerupgradethreshold

            Get the ROUTER_UPGRADE_THRESHOLD value.

            > routerupgradethreshold
            16
            Done

        routerupgradethreshold <threshold>

            Set the ROUTER_UPGRADE_THRESHOLD value.

            > routerupgradethreshold 16
            Done
        """
        self.handle_property(line, SPINEL.PROP_THREAD_ROUTER_UPGRADE_THRESHOLD)

    def do_routerdowngradethreshold(self, line):
        """
        routerdowngradethreshold

            Get the ROUTER_DOWNGRADE_THRESHOLD value.

            > routerdowngradethreshold
            16
            Done

        routerdowngradethreshold <threshold>

            Set the ROUTER_DOWNGRADE_THRESHOLD value.

            > routerdowngradethreshold 16
            Done
        """
        self.handle_property(line,
                             SPINEL.PROP_THREAD_ROUTER_DOWNGRADE_THRESHOLD)

    def do_scan(self, _line):
        """
        scan [channel]

            Perform an IEEE 802.15.4 Active Scan.

              channel: The channel to scan on. If no channel is provided,
              the active scan will cover all valid channels.

            > scan
            | J | Network Name     | Extended PAN     | PAN  | MAC Address      | Ch | dBm | LQI |
            +---+------------------+------------------+------+------------------+----+-----+-----+
            | 0 | OpenThread       | dead00beef00cafe | ffff | f1d92a82c8d8fe43 | 11 | -20 |   0 |
        Done
        """
        # Initial mock-up of scan
        self.handle_property("15", SPINEL.PROP_MAC_SCAN_MASK)
        self.handle_property("4", SPINEL.PROP_MAC_SCAN_PERIOD, 'H')
        self.handle_property("1", SPINEL.PROP_MAC_SCAN_STATE)
        time.sleep(5)
        self.handle_property("", SPINEL.PROP_MAC_SCAN_BEACON, 'U')

    def complete_thread(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for thread command. """
        map_sub_commands = ('start', 'stop')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_thread(self, line):
        """
        thread start

            Enable Thread protocol operation and attach to a Thread network.

            > thread start
            Done

        thread stop

            Disable Thread protocol operation and detach from a Thread network.

            > thread stop
            Done
        """
        map_arg_value = {
            0: "stop",
            1: "start",
        }

        map_arg_name = {
            "stop": "0",
            "start": "1",
        }

        if line:
            try:
                # remap string state names to integer
                line = map_arg_name[line]
            except:
                print("Error")
                return

        result = self.prop_get_or_set_value(SPINEL.PROP_NET_STACK_UP, line)
        if result != None:
            if not line:
                print(map_arg_value[result])
            print("Done")
        else:
            print("Error")

    def do_state(self, line):
        """
        state
        """
        map_arg_value = {
            0: "detached",
            1: "child",
            2: "router",
            3: "leader",
        }

        map_arg_name = {
            "disabled": "0",
            "detached": "0",
            "child": "1",
            "router": "2",
            "leader": "3",
        }

        if line:
            try:
                # remap string state names to integer
                line = map_arg_name[line]
            except:
                print("Error")
                return

        result = self.prop_get_or_set_value(SPINEL.PROP_NET_ROLE, line)
        if result != None:
            if not line:
                state = map_arg_value[result]
                # TODO: if state="disabled": get NET_STATE to determine
                #       whether "disabled" or "detached"
                print(state)
            print("Done")
        else:
            print("Error")

    def do_txpower(self, line):
        """
        txpower

            Get the transmit power in dBm.

            > txpower
            0
            Done

        txpower <txpower>

            Set the transmit power in dBm.

            > txpower -10
            Done
        """
        self.handle_property(line, SPINEL.PROP_PHY_TX_POWER, mixed_format='b')

    def do_version(self, line):
        """
        version

            Print the build version information.

            > version
            OPENTHREAD/gf4f2f04; Jul  1 2016 17:00:09
            Done
        """
        self.handle_property(line, SPINEL.PROP_NCP_VERSION, 'U')

    def complete_macfilter(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for macfilter command. """
        #TODO: autocomplete the secondary sub commands
        #for 'addr': 'disable', 'denylist', 'allowlist', 'add', 'remove', 'clear'
        #for 'rss' : 'add', 'remove', 'clear'
        map_sub_commands = ('addr', 'rss')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_mac(self, line):
        """
        mac

            Mac related commands.

        mac retries direct

            Get the number of transmit retries on the MAC layer.

            > mac retries direct
            3
            Done

        mac retries direct <number>

            Set the number of direct transmit retries on the MAC layer.

            > mac retries direct 10
            Done

        mac retries indirect

            Get the number of indirect transmit retries on the MAC layer.

            > mac retries indirect
            0
            Done

        mac retries indirect <number>

            Set the number of indirect transmit retries on the MAC layer.

            > mac retries indirect 5
            Done

        mac ccathreshold

            Get the CCA ED Threshold in dBm.

            > mac ccathreshold
            -10
            Done

        mac ccathreshold -70

            Set the CCA ED Threshold in dBm.

            > mac ccathreshold -70
            Done
        """
        params = line.split(" ")
        prop = None

        if params[0] == "retries" and len(params) > 1:
            if params[1] == "direct":
                prop = SPINEL.PROP_MAC_MAX_RETRY_NUMBER_DIRECT
            elif params[1] == "indirect":
                prop = SPINEL.PROP_MAC_MAX_RETRY_NUMBER_INDIRECT
            value = params[2] if len(params) == 3 else None

        elif params[0] == "ccathreshold" and len(params) > 0:
            prop = SPINEL.PROP_PHY_CCA_THRESHOLD
            value = None
            if len(params) == 2:
                value = int(params[1])
                self.prop_set(prop, value, mixed_format='b')
                return

        self.handle_property(value, prop)

    def do_macfilter(self, line):
        """
        macfilter

           List the macfilter status, including address and received signal strength filter settings.

           > macfilter
           Allowlist
           Done

        macfilter addr

            List the address filter status.

            > macfilter addr
            Allowlist
            Done

        macfilter addr disable
            Disable address filter mode.

            > macfilter addr disable
            Done

        macfilter addr allowlist
            Enable allowlist address filter mode.

            > macfilter addr allowlist
            Done

        macfilter addr denylist
            Enable denylist address filter mode.

            > macfilter addr denylist
            Done

        macfilter addr add <extaddr> [rssi]

            Add an IEEE 802.15.4 Extended Address to the address filter.

            > macfilter addr add dead00beef00cafe -85
            Done

            > macfilter addr add dead00beef00caff
            Done

        macfilter addr remove <extaddr>

            Remove an IEEE 802.15.4 Extended Address from the address filter.

            > macfilter addr remove dead00beef00caff
            Done


        macfilter addr clear

            Clear all entries from the address filter.

            > macfilter addr clear
            Done

        macfilter rss

            List the rss filter status.

            > macfilter rss
            Done

        macfilter rss add <extaddr> <rssi>

            Set the received signal strength for the messages from the IEEE802.15.4 Extended Address.
            If extaddr is \*, default received signal strength for all received messages would be set.

            > macfilter rss add * -50
            Done

            > macfilter rss add 0f6127e33af6b404 -85
            Done

        macfilter rss remove <extaddr>

            Removes the received signal strength or received link quality setting on the Extended Address.
            If extaddr is \*, default received signal strength or link quality for all received messages would be unset.

            > macfilter rss remove *
            Done

            > macfilter rss remove 0f6127e33af6b404

        macfilter rss clear

            Clear all the the received signal strength.

            > macfilter rss clear
        """

        map_arg_value = {
            0: "Disabled",
            1: "Allowlist",
            2: "Denylist",
        }

        params = line.split(" ")

        if params[0] == "":
            mode = 0
            value = self.prop_get_value(SPINEL.PROP_MAC_ALLOWLIST_ENABLED)
            if value == 1:
                mode = 1
            else:
                value = self.prop_get_value(SPINEL.PROP_MAC_DENYLIST_ENABLED)
                if value == 1:
                    mode = 2

            print(map_arg_value[mode])

            # TODO: parse and show the content of entries
            value = self.prop_get_value(SPINEL.PROP_MAC_ALLOWLIST)
            value = self.prop_get_value(SPINEL.PROP_MAC_FIXED_RSS)

        if params[0] == "addr":
            if len(params) == 1:
                mode = 0
                value = self.prop_get_value(SPINEL.PROP_MAC_ALLOWLIST_ENABLED)
                if value == 1:
                    mode = 1
                else:
                    value = self.prop_get_value(
                        SPINEL.PROP_MAC_DENYLIST_ENABLED)
                    if value == 1:
                        mode = 2

                print(map_arg_value[mode])
                # TODO: parse and show the content of entries
                value = self.prop_get_value(SPINEL.PROP_MAC_ALLOWLIST)

            elif params[1] == "allowlist":
                self.prop_set(SPINEL.PROP_MAC_ALLOWLIST_ENABLED, '1')
                return

            elif params[1] == "denylist":
                self.prop_set(SPINEL.PROP_MAC_DENYLIST_ENABLED, '1')
                return

            elif params[1] == "disable":
                self.prop_set(SPINEL.PROP_MAC_ALLOWLIST_ENABLED, '0')
                return

            elif params[1] == "add":
                arr = util.hex_to_bytes(params[2])
                try:
                    rssi = int(params[3])
                except:
                    rssi = SPINEL.RSSI_OVERRIDE

                arr += struct.pack('b', rssi)
                self.prop_insert_value(SPINEL.PROP_MAC_ALLOWLIST, arr,
                                       str(len(arr)) + 's')

            elif params[1] == "remove":
                arr = util.hex_to_bytes(params[2])
                self.prop_remove_value(SPINEL.PROP_MAC_ALLOWLIST, arr,
                                       str(len(arr)) + 's')
            elif params[1] == "clear":
                self.prop_set_value(SPINEL.PROP_MAC_ALLOWLIST, b'', '0s')

        elif params[0] == "rss":
            if len(params) == 1:
                # TODO: parse and show the content of entries
                value = self.prop_get_value(SPINEL.PROP_MAC_FIXED_RSS)

            elif params[1] == "add":
                if params[2] == "*":
                    arr = b''
                else:
                    arr = util.hex_to_bytes(params[2])
                rssi = int(params[3])
                arr += struct.pack('b', rssi)
                self.prop_insert_value(SPINEL.PROP_MAC_FIXED_RSS, arr,
                                       str(len(arr)) + 's')

            elif params[1] == "remove":
                if params[2] == "*":
                    arr = b''
                else:
                    arr = util.hex_to_bytes(params[2])
                self.prop_remove_value(SPINEL.PROP_MAC_FIXED_RSS, arr,
                                       str(len(arr)) + 's')

            elif params[1] == "clear":
                self.prop_set_value(SPINEL.PROP_MAC_FIXED_RSS, b'', '0s')

        print("Done")

    def do_ncpll64(self, line):
        """ Display the link local IPv6 address. """
        self.handle_property(line, SPINEL.PROP_IPV6_LL_ADDR, '6')

    def do_ncpml64(self, line):
        """ Display the mesh local IPv6 address. """
        self.handle_property(line, SPINEL.PROP_IPV6_ML_ADDR, '6')

    def do_ncpraw(self, line):
        """ Enable MAC raw stream. """
        self.handle_property(line, SPINEL.PROP_MAC_RAW_STREAM_ENABLED, 'B')

    def do_ncpfilter(self, line):
        """
        Set MAC filter mode:

        0 = MAC_FILTER_MODE_NORMAL	Normal MAC filtering is in place.
        1 = MAC_FILTER_MODE_PROMISCUOUS	All MAC packets matching network are passed up the stack.
        2 = MAC_FILTER_MODE_MONITOR	All decoded MAC packets are passed up the stack.
        """
        self.handle_property(line, SPINEL.PROP_MAC_FILTER_MODE, 'B')

    def complete_ncptun(self, text, _line, _begidx, _endidx):
        """ Subcommand completion handler for ncp-tun command. """
        map_sub_commands = ('up', 'down', 'add', 'remove', 'ping')
        return [i for i in map_sub_commands if i.startswith(text)]

    def do_ncptun(self, line):
        """
        ncp-tun

            Control sideband tunnel interface.

        ncp-tun up

            Bring up Thread TUN interface.

            > ncp-tun up
            Done

        ncp-tun down

            Bring down Thread TUN interface.

            > ncp-tun down
            Done

        ncp-tun add <ipaddr>

            Add an IPv6 address to the Thread TUN interface.

            > ncp-tun add 2001::dead:beef:cafe
            Done

        ncp-tun del <ipaddr>

            Delete an IPv6 address from the Thread TUN interface.

            > ncp-tun del 2001::dead:beef:cafe
            Done

        ncp-tun ping <ipaddr> [size] [count] [interval]

            Send an ICMPv6 Echo Request.

            > ncp-tun ping fdde:ad00:beef:0:558:f56b:d688:799
            16 bytes from fdde:ad00:beef:0:558:f56b:d688:799: icmp_seq=1 hlim=64 time=28ms
        """
        params = line.split(" ")

        num = len(params)
        if num > 1:
            ipaddr = params[1]
            prefix = ipaddress.IPv6Interface(str(ipaddr))
            _arr = prefix.ip.packed

        if params[0] == "":
            pass

        elif params[0] == "add":
            if self.tun_if:
                self.tun_if.addr_add(ipaddr)

        elif params[0] == "remove":
            if self.tun_if:
                self.tun_if.addr_del(ipaddr)

        elif params[0] == "up":
            if os.geteuid() == 0:
                self.tun_if = TunInterface(self.nodeid)
            else:
                print("Warning: superuser required to start tun interface.")

        elif params[0] == "down":
            if self.tun_if:
                self.tun_if.close()
            self.tun_if = None

        elif params[0] == "ping":
            # Use tunnel to send ping
            size = "56"
            count = "1"
            _interval = "1"
            if len(params) > 1:
                size = params[1]
            if len(params) > 2:
                count = params[2]
            if len(params) > 3:
                _interval = params[3]

            if self.tun_if:
                self.tun_if.ping6(" -c " + count + " -s " + size + " " + ipaddr)

        print("Done")

    def do_diag(self, line):
        """
        Follows "mfg" command.
        """
        self.do_mfg(line)

    def _notify_simulator(self):
        """
        notify the simulator that there are no more UART data for the current command.
        """
        OT_SIM_EVENT_POSTCMD = 4

        message = struct.pack('=QBHB', 0, OT_SIM_EVENT_POSTCMD, 1,
                              int(self.nodeid))
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind(self._addr)
        sock.sendto(message, self._simulator_addr)
        sock.close()

    def postcmd(self, stop, line):
        if self.VIRTUAL_TIME:
            self._notify_simulator()
        return stop


def parse_args():
    """" Send spinel commands to initialize sniffer node. """
    args = sys.argv[1:]

    opt_parser = optparse.OptionParser(usage=optparse.SUPPRESS_USAGE)
    opt_parser.add_option("-u",
                          "--uart",
                          action="store",
                          dest="uart",
                          type="string")
    opt_parser.add_option("-b",
                          "--baudrate",
                          action="store",
                          dest="baudrate",
                          type="int",
                          default=DEFAULT_BAUDRATE)
    opt_parser.add_option("--rtscts",
                          action="store_true",
                          dest="rtscts",
                          default=False),
    opt_parser.add_option("-p",
                          "--pipe",
                          action="store",
                          dest="pipe",
                          type="string")
    opt_parser.add_option("-s",
                          "--socket",
                          action="store",
                          dest="socket",
                          type="string")
    opt_parser.add_option("-n",
                          "--nodeid",
                          action="store",
                          dest="nodeid",
                          type="string",
                          default="1")
    opt_parser.add_option("-q", "--quiet", action="store_true", dest="quiet")
    opt_parser.add_option("-v",
                          "--verbose",
                          action="store_true",
                          dest="verbose")
    opt_parser.add_option("-d",
                          "--debug",
                          action="store",
                          dest="debug",
                          type="int",
                          default=CONFIG.DEBUG_ENABLE)
    opt_parser.add_option("--vendor-path",
                          action="store",
                          dest="vendor_path",
                          type="string")

    return opt_parser.parse_args(args)


def main():
    """ Top-level main for spinel-cli tool. """
    (options, remaining_args) = parse_args()

    if options.debug:
        CONFIG.debug_set_level(options.debug)

    # Obtain the vendor module path, if provided
    if not options.vendor_path:
        options.vendor_path = os.environ.get("SPINEL_VENDOR_PATH")

    if options.vendor_path:
        options.vendor_path = os.path.abspath(options.vendor_path)
        vendor_path, vendor_module = os.path.split(options.vendor_path)
        sys.path.insert(0, vendor_path)
    else:
        vendor_module = "vendor"

    # Set default stream to pipe
    stream_type = 'p'
    stream_descriptor = "../../examples/apps/ncp/ot-ncp-ftd " + options.nodeid

    if options.uart:
        stream_type = 'u'
        stream_descriptor = options.uart
    elif options.socket:
        stream_type = 's'
        stream_descriptor = options.socket
    elif options.pipe:
        stream_type = 'p'
        stream_descriptor = options.pipe
        if options.nodeid:
            stream_descriptor += " " + str(options.nodeid)
    else:
        if len(remaining_args) > 0:
            stream_descriptor = " ".join(remaining_args)

    stream = StreamOpen(stream_type, stream_descriptor, options.verbose,
                        options.baudrate, options.rtscts)
    try:
        vendor_ext = importlib.import_module(vendor_module + '.vendor')
        cls = type(vendor_ext.VendorSpinelCliCmd.__name__,
                   (SpinelCliCmd, vendor_ext.VendorSpinelCliCmd), {})
        shell = cls(stream, nodeid=options.nodeid, vendor_module=vendor_module)
    except ImportError:
        shell = SpinelCliCmd(stream,
                             nodeid=options.nodeid,
                             vendor_module=vendor_module)

    try:
        shell.cmdloop()
    except KeyboardInterrupt:
        CONFIG.LOGGER.info('\nCTRL+C Pressed')

    if shell.wpan_api:
        shell.wpan_api.stream.close()


if __name__ == "__main__":
    main()



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Module providing a specific vendor commands.
"""

from vendor.const import VENDOR_SPINEL


class VendorSpinelCliCmd():
    """
    Extended Vendor Spinel Cli with vendor hooks commands.
    INPUT:
            spinel-cli > vendor help
    OUTPUT:
            Available vendor commands:
            ==============================================
            help
    """
    vendor_command_names = ['help']

    def do_vendor(self, line):
        params = line.split(" ")
        if params[0] == 'help':
            self.print_topics("\nAvailable vendor commands:",
                              VendorSpinelCliCmd.vendor_command_names, 15, 30)



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Unittest for spinel.hdlc module.
"""

import unittest
import binascii

from spinel.hdlc import Hdlc


class TestHdlc(unittest.TestCase):
    """ Unittest class for spinel.hdlc.Hdlc class. """

    VECTOR = {
        # Data    HDLC Encoded
        "810243": "7e810243d3d37e",
        "8103367e7d": "7e8103367d5e7d5d6af97e",
    }

    def test_hdlc_encode(self):
        """ Unit test for Hdle.encode method. """
        hdlc = Hdlc(None)
        for in_hex, out_hex in self.VECTOR.iteritems():
            in_binary = binascii.unhexlify(in_hex)
            out_binary = hdlc.encode(in_binary)
            #print "inHex = "+binascii.hexlify(in_binary)
            #print "outHex = "+binascii.hexlify(out_binary)
            self.failUnless(out_hex == binascii.hexlify(out_binary))

    def test_hdlc_decode(self):
        """ Unit test for Hdle.decode method. """
        pass



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
""" Unittest for spinel.codec module. """

import time
import unittest

from spinel.const import SPINEL
from spinel.codec import WpanApi
from spinel.test_stream import MockStream


class TestCodec(unittest.TestCase):
    """ Unit TestCase class for spinel.codec.SpinelCodec class. """

    # Tests parsing and format demuxing of various properties with canned
    # values.
    VECTOR = {
        SPINEL.PROP_MAC_15_4_PANID: 65535,
        SPINEL.PROP_NCP_VERSION: "OPENTHREAD",
        SPINEL.PROP_NET_ROLE: 0,
        SPINEL.PROP_NET_KEY_SEQUENCE_COUNTER: 5,
        SPINEL.PROP_NET_NETWORK_NAME: "OpenThread",
        SPINEL.PROP_THREAD_MODE: 0xF,
    }

    def test_prop_get(self):
        """ Unit test of SpinelCodec.prop_get_value. """

        mock_stream = MockStream({
            # Request:  Response
            "810236": "810636ffff",  # get panid = 65535
            "810243": "81064300",  # get state = detached
            "81025e": "81065e0f",  # mode = 0xF
            "810202": "8106024f50454e54485245414400",  # get version
            "810247": "81064705000000",  # get keysequence
            "810244": "8106444f70656e54687265616400",  # get networkname
        })
        nodeid = 1
        use_hdlc = False
        wpan_api = WpanApi(mock_stream, nodeid, use_hdlc)

        for prop_id, truth_value in self.VECTOR.iteritems():
            value = wpan_api.prop_get_value(prop_id)
            self.failUnless(value == truth_value)

    def cb_test_callback(self, prop, value, tid):
        self.test_callback_pass = True

    def test_callback(self):
        """ Unit test of WpanApi.callback_register. """

        vector = [
            "800672340060000000000c3a40fe80000000000000020d6f00055715d3fddead00beef0000cd9bb7814c5619ea8100b0ca00000000267fc789"  # PROP_STREAM_NET
        ]

        mock_stream = MockStream({})
        nodeid = 1
        use_hdlc = False
        wpan_api = WpanApi(mock_stream, nodeid, use_hdlc)

        self.test_callback_pass = False
        wpan_api.callback_register(SPINEL.PROP_STREAM_NET,
                                   self.cb_test_callback)

        for pkt in vector:
            mock_stream.write_child_hex(pkt)
            time.sleep(0.1)

        self.failUnless(self.test_callback_pass)



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Module providing a Vendor property handlers.
"""

from spinel.codec import SpinelCodec
from vendor.const import VENDOR_SPINEL


class VendorSpinelPropertyHandler(SpinelCodec):
    """
    Class to extend Spinel property Handler with new methods.
    Methods define parsers for Vendor Hooks for exapmle:
        `def VENDOR_HOOK_PROPERTY(self, _wpan_api, payload): return self.parse_C(payload)`
    """
    pass


WPAN_PROP_HANDLER = VendorSpinelPropertyHandler()

# Parameter to extend SPINEL_PREP_DISPATCH with Vendor properties for example:
#   `VENDOR_SPINEL_PROP_DISPATCH = {VENDOR_SPINEL.PROP_VENDOR_HOOK: WPAN_PROP_HANDLER.VENDOR_HOOK_PROPERTY}`
VENDOR_SPINEL_PROP_DISPATCH = {}



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
""" Module to provide codec utilities for .pcap formatters. """

import struct

PCAP_MAGIC_NUMBER = 0xA1B2C3D4
PCAP_VERSION_MAJOR = 2
PCAP_VERSION_MINOR = 4

# https://www.tcpdump.org/linktypes.html
DLT_IEEE802_15_4_WITHFCS = 195
DLT_IEEE802_15_4_TAP = 283

# Refer to the IEEE 802.15.4 TAP Link Type Specification on
# https://github.com/jkcko/ieee802.15.4-tap
# Default length of TAP Header and channel TLV
TLVS_LENGTH_DEFAULT = 12
CHANNEL_TYPE = 3
CHANNEL_LEN = 3
CHANNEL_PAGE = 0

# FCS TLV (optional, depending on `--crc`)
FCS_TYPE = 0
FCS_LEN = 1
FCS_16bitCRC = 1

# RSSI TLV and LQI TLV (optional, depending on `--rssi`)
RSS_TYPE = 1
RSS_LEN = 4
LQI_TYPE = 10
LQI_LEN = 1


def crc(s):
    # Some chips do not transmit the CRC, here we recalculate the CRC.

    crc = 0
    # remove the last 2 bytes
    for c in s[:-2]:
        q = (crc ^ c) & 0x0F
        crc = (crc >> 4) ^ (q * 0x1081)
        q = (crc ^ (c >> 4)) & 0x0F
        crc = (crc >> 4) ^ (q * 0x1081)

    # lsb
    s[-2] = 0xFF & (crc >> 0)
    # msb
    s[-1] = 0xFF & (crc >> 8)
    return s


class PcapCodec(object):
    """ Utility class for .pcap formatters. """

    @classmethod
    def encode_header(cls, dlt):
        """ Returns a pcap file header. """
        cls._dlt = dlt
        return struct.pack("<LHHLLLL", PCAP_MAGIC_NUMBER, PCAP_VERSION_MAJOR,
                           PCAP_VERSION_MINOR, 0, 0, 256, cls._dlt)

    @classmethod
    def encode_frame(cls,
                     frame,
                     sec,
                     usec,
                     options_rssi,
                     options_crc,
                     metadata=None):
        """ Returns a pcap encapsulation of the given frame. """
        # write frame pcap header
        TLVs_length = TLVS_LENGTH_DEFAULT

        frame = bytearray(frame)

        if options_crc:
            frame = crc(frame)
            TLVs_length += 8

        if options_rssi:
            if cls._dlt == DLT_IEEE802_15_4_TAP:
                TLVs_length += 16
            else:
                # TI style FCS format: replace the last two bytes (should be FCS) with RSSI and LQI and always
                # assume FCS right
                frame[-1] = metadata[0] & 0xFF
                frame[-2] = metadata[3][1] & 0xFF

        if cls._dlt == DLT_IEEE802_15_4_TAP:
            length = len(frame) + TLVs_length
        else:
            length = len(frame)

        pcap_frame = struct.pack("<LLLL", sec, usec, length, length)

        if cls._dlt == DLT_IEEE802_15_4_TAP:
            # Append TLVs according to 802.15.4 TAP specification:
            # https://github.com/jkcko/ieee802.15.4-tap
            pcap_frame += struct.pack('<HH', 0, TLVs_length)
            pcap_frame += struct.pack('<HHHH', CHANNEL_TYPE, CHANNEL_LEN,
                                      metadata[3][0], CHANNEL_PAGE)
            if options_rssi:
                pcap_frame += struct.pack('<HHf', RSS_TYPE, RSS_LEN,
                                          metadata[0])
                pcap_frame += struct.pack('<HHI', LQI_TYPE, LQI_LEN,
                                          metadata[3][1])
            if options_crc:
                pcap_frame += struct.pack('<HHI', FCS_TYPE, FCS_LEN,
                                          FCS_16bitCRC)

        pcap_frame += frame
        return pcap_frame



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Module providing a Spienl coder / decoder class.
"""

import binascii
import time
import logging
import threading
import traceback
import queue
import importlib

from struct import pack
from struct import unpack
from collections import namedtuple
from collections import defaultdict

import ipaddress

import spinel.util as util
import spinel.config as CONFIG
from spinel.const import kThread
from spinel.const import SPINEL
from spinel.const import SPINEL_LAST_STATUS_MAP
from spinel.hdlc import Hdlc

FEATURE_USE_HDLC = 1
FEATURE_USE_SLACC = 1

TIMEOUT_PROP = 2

#=========================================
#   SpinelCodec
#=========================================

#  0: DATATYPE_NULL
#'.': DATATYPE_VOID: Empty data type. Used internally.
#'b': DATATYPE_BOOL: Boolean value. Encoded in 8-bits as either 0x00 or 0x01.
#                    All other values are illegal.
#'C': DATATYPE_UINT8: Unsigned 8-bit integer.
#'c': DATATYPE_INT8: Signed 8-bit integer.
#'S': DATATYPE_UINT16: Unsigned 16-bit integer. (Little-endian)
#'s': DATATYPE_INT16: Signed 16-bit integer. (Little-endian)
#'L': DATATYPE_UINT32: Unsigned 32-bit integer. (Little-endian)
#'l': DATATYPE_INT32: Signed 32-bit integer. (Little-endian)
#'i': DATATYPE_UINT_PACKED: Packed Unsigned Integer. (See section 7.2)
#'6': DATATYPE_IPv6ADDR: IPv6 Address. (Big-endian)
#'E': DATATYPE_EUI64: EUI-64 Address. (Big-endian)
#'e': DATATYPE_EUI48: EUI-48 Address. (Big-endian)
#'D': DATATYPE_DATA: Arbitrary Data. (See section 7.3)
#'d': DATATYPE_DATA_WLEN: Arbitrary Data with Prepended Length. (See section 7.3)
#'U': DATATYPE_UTF8: Zero-terminated UTF8-encoded string.
#'t': DATATYPE_STRUCT: Structured datatype. Compound type. Length prepended. (See section 7.4)
#'A': DATATYPE_ARRAY: Array of datatypes. Compound type. (See section 7.5)


class SpinelCodec(object):
    """ A general coder / decoder class for Spinel protocol. """

    @classmethod
    def parse_b(cls, payload):
        return unpack("<B", payload[:1])[0]

    @classmethod
    def parse_c(cls, payload):
        return unpack("<b", payload[:1])[0]

    @classmethod
    def parse_C(cls, payload):
        return unpack("<B", payload[:1])[0]

    @classmethod
    def parse_s(cls, payload):
        return unpack("<h", payload[:2])[0]

    @classmethod
    def parse_S(cls, payload):
        return unpack("<H", payload[:2])[0]

    @classmethod
    def parse_l(cls, payload):
        return unpack("<l", payload[:4])[0]

    @classmethod
    def parse_L(cls, payload):
        return unpack("<L", payload[:4])[0]

    @classmethod
    def parse_X(cls, payload):
        return unpack("<Q", payload[:8])[0]

    @classmethod
    def parse_6(cls, payload):
        return payload[:16]

    @classmethod
    def parse_E(cls, payload):
        return payload[:8]

    @classmethod
    def parse_e(cls, payload):
        return payload[:6]

    @classmethod
    def parse_U(cls, payload):
        payload = payload.decode('utf-8')
        nullchar = '\0'

        if payload.find(nullchar) >= 0:
            return payload[:payload.index(nullchar)]  # strip null
        else:
            return payload

    @classmethod
    def parse_D(cls, payload):
        return payload

    @classmethod
    def parse_d(cls, payload):
        return payload[2:2 + unpack("<H", payload[:2])[0]]

    @classmethod
    def parse_i(cls, payload):
        """ Decode EXI integer format. """
        value = 0
        value_len = 0
        value_mul = 1

        while value_len < 4:
            byte = payload[value_len]

            value += (byte & 0x7F) * value_mul
            if byte < 0x80:
                break
            value_mul *= 0x80
            value_len += 1

        return (value, value_len + 1)

    @classmethod
    def parse_i_len(cls, payload):
        """ Decode length of EXI integer format. """
        return cls.parse_i(payload)[1]

    @classmethod
    def index_of_ending_brace(cls, spinel_format, idx):
        """ Determines the index of the matching closing brace. """
        count = 1
        orig_idx = idx
        while count > 0 and idx < len(spinel_format) - 1:
            idx += 1
            if spinel_format[idx] == ')':
                count -= 1
            elif spinel_format[idx] == '(':
                count += 1
        if count != 0:
            raise ValueError('Unbalanced parenthesis in format string "' +
                             spinel_format + '", idx=' + idx)
        return idx

    @classmethod
    def parse_field(cls, payload, spinel_format):
        map_decode = {
            'b': cls.parse_b,
            'c': cls.parse_c,
            'C': cls.parse_C,
            's': cls.parse_s,
            'S': cls.parse_S,
            'L': cls.parse_L,
            'l': cls.parse_l,
            '6': cls.parse_6,
            'X': cls.parse_X,
            'E': cls.parse_E,
            'e': cls.parse_e,
            'U': cls.parse_U,
            'D': cls.parse_D,
            'd': cls.parse_d,
            'i': cls.parse_i,
        }
        try:
            return map_decode[spinel_format[0]](payload)
        except KeyError:
            print(traceback.format_exc())
            return None

    @classmethod
    def get_payload_size(cls, payload, spinel_format):
        map_lengths = {
            'b': 1,
            'c': 1,
            'C': 1,
            's': 2,
            'S': 2,
            'l': 4,
            'L': 4,
            '6': 16,
            'X': 8,
            'E': 8,
            'e': 6,
        }

        result = 0

        idx = 0
        while idx < len(spinel_format):
            format = spinel_format[idx]

            if format == 't':
                if spinel_format[idx + 1] != '(':
                    raise ValueError('Invalid structure format')

                struct_end = cls.index_of_ending_brace(spinel_format, idx + 1)

                result += 2 + cls.parse_S(payload[result:])
                idx = struct_end + 1

            elif format == 'd':
                result += 2 + cls.parse_S(payload[result:])
                idx += 1

            elif format == 'D' or format == 'A':
                if idx != len(spinel_format) - 1:
                    raise ValueError('Invalid type syntax for "' + format +
                                     '", must go at end of format string')
                result = len(payload)
                idx += 1

            elif format == 'U':
                result += payload[result:].index(0) + 1
                idx += 1

            elif format == 'i':
                result += cls.parse_i_len(payload[result:])
                idx += 1

            else:
                result += map_lengths[format]
                idx += 1

        return result

    @classmethod
    def parse_fields(cls, payload, spinel_format):
        result = []

        idx = 0
        while idx < len(spinel_format):
            format = spinel_format[idx]

            if format == 'A':
                if spinel_format[idx + 1] != '(':
                    raise ValueError('Invalid structure format')

                array_end = cls.index_of_ending_brace(spinel_format, idx + 1)

                array_format = spinel_format[idx + 2:array_end]

                array = []
                while len(payload):
                    array.append(cls.parse_fields(payload, array_format))
                    payload = payload[cls.
                                      get_payload_size(payload, array_format):]

                result.append(tuple(array))
                idx = array_end + 1

            elif format == 't':
                if spinel_format[idx + 1] != '(':
                    raise ValueError('Invalid structure format')

                struct_end = cls.index_of_ending_brace(spinel_format, idx + 1)

                struct_format = spinel_format[idx + 2:struct_end]
                struct_len = cls.parse_S(payload)

                result.append(
                    cls.parse_fields(payload[2:struct_len + 2], struct_format))
                payload = payload[struct_len + 2:]

                idx = struct_end + 1

            else:
                result.append(cls.parse_field(payload, format))
                payload = payload[cls.get_payload_size(payload, format):]

                idx += 1

        return tuple(result)

    @classmethod
    def encode_i(cls, data):
        """ Encode EXI integer format. """
        result = bytes()
        while data:
            value = data & 0x7F
            data >>= 7
            if data:
                value |= 0x80
            result = result + pack("<B", value)
        return result

    @classmethod
    def encode_b(cls, value):
        return pack('B', value)

    @classmethod
    def encode_c(cls, value):
        return pack('B', value)

    @classmethod
    def encode_C(cls, value):
        return pack('B', value)

    @classmethod
    def encode_s(cls, value):
        return pack('<h', value)

    @classmethod
    def encode_S(cls, value):
        return pack('<H', value)

    @classmethod
    def encode_l(cls, value):
        return pack('<l', value)

    @classmethod
    def encode_L(cls, value):
        return pack('<L', value)

    @classmethod
    def encode_6(cls, value):
        return value[:16]

    @classmethod
    def encode_E(cls, value):
        return value[:8]

    @classmethod
    def encode_e(cls, value):
        return value[:6]

    @classmethod
    def encode_U(cls, value):
        return value + '\0'

    @classmethod
    def encode_D(cls, value):
        return value

    @classmethod
    def encode_d(cls, value):
        return cls.encode_S(len(value)) + value

    @classmethod
    def encode_field(cls, code, value):
        map_encode = {
            'b': cls.encode_b,
            'c': cls.encode_c,
            'C': cls.encode_C,
            's': cls.encode_s,
            'S': cls.encode_S,
            'L': cls.encode_L,
            'l': cls.encode_l,
            '6': cls.encode_6,
            'E': cls.encode_E,
            'e': cls.encode_e,
            'U': cls.encode_U,
            'D': cls.encode_D,
            'd': cls.encode_d,
            'i': cls.encode_i,
        }
        try:
            return map_encode[code](value)
        except KeyError:
            print(traceback.format_exc())
            return None

    def next_code(self, spinel_format):
        code = spinel_format[0]
        spinel_format = spinel_format[1:]
        # TODO: Handle T() and A()
        return code, spinel_format

    def encode_fields(self, spinel_format, *fields):
        packed = bytes()
        for field in fields:
            code, spinel_format = self.next_code(spinel_format)
            if not code:
                break
            packed += self.encode_field(code, field)
        return packed

    def encode_packet(self,
                      command_id,
                      payload=bytes(),
                      tid=SPINEL.HEADER_DEFAULT):
        """ Encode the given payload as a Spinel frame. """
        header = pack(">B", tid)
        cmd = self.encode_i(command_id)
        pkt = header + cmd + payload
        return pkt


#=========================================


class SpinelPropertyHandler(SpinelCodec):

    def LAST_STATUS(self, _, payload):
        return self.parse_i(payload)[0]

    def PROTOCOL_VERSION(self, _wpan_api, payload):
        pass

    def NCP_VERSION(self, _, payload):
        return self.parse_U(payload)

    def INTERFACE_TYPE(self, _, payload):
        return self.parse_i(payload)[0]

    def VENDOR_ID(self, _, payload):
        return self.parse_i(payload)[0]

    def CAPS(self, _wpan_api, payload):
        return self.parse_fields(payload, 'A(i)')

    def INTERFACE_COUNT(self, _, payload):
        return self.parse_C(payload)

    def POWER_STATE(self, _, payload):
        return self.parse_C(payload)

    def HWADDR(self, _, payload):
        return self.parse_E(payload)

    def LOCK(self, _, payload):
        return self.parse_b(payload)

    def HBO_MEM_MAX(self, _, payload):
        return self.parse_L(payload)

    def HBO_BLOCK_MAX(self, _, payload):
        return self.parse_S(payload)

    def PHY_ENABLED(self, _, payload):
        return self.parse_b(payload)

    def PHY_CHAN(self, _, payload):
        return self.parse_C(payload)

    def PHY_CHAN_SUPPORTED(self, _wpan_api, payload):
        pass

    def PHY_FREQ(self, _, payload):
        return self.parse_L(payload)

    def PHY_CCA_THRESHOLD(self, _, payload):
        return self.parse_c(payload)

    def PHY_TX_POWER(self, _, payload):
        return self.parse_c(payload)

    def PHY_RSSI(self, _, payload):
        return self.parse_c(payload)

    def MAC_SCAN_STATE(self, _, payload):
        return self.parse_C(payload)

    def MAC_SCAN_MASK(self, _, payload):
        return self.parse_U(payload)

    def MAC_SCAN_PERIOD(self, _, payload):
        return self.parse_S(payload)

    def MAC_SCAN_BEACON(self, _, payload):
        return self.parse_U(payload)

    def MAC_15_4_LADDR(self, _, payload):
        return self.parse_E(payload)

    def MAC_15_4_SADDR(self, _, payload):
        return self.parse_S(payload)

    def MAC_15_4_PANID(self, _, payload):
        return self.parse_S(payload)

    def MAC_FILTER_MODE(self, _, payload):
        return self.parse_C(payload)

    def MAC_RAW_STREAM_ENABLED(self, _, payload):
        return self.parse_b(payload)

    def MAC_ALLOWLIST(self, _, payload):
        pass

    def MAC_ALLOWLIST_ENABLED(self, _, payload):
        return self.parse_b(payload)

    def MAC_DENYLIST(self, _, payload):
        pass

    def MAC_DENYLIST_ENABLED(self, _, payload):
        return self.parse_b(payload)

    def MAC_FIXED_RSS(self, _, payload):
        pass

    def MAC_MAX_RETRY_NUMBER_DIRECT(self, _, payload):
        return self.parse_C(payload)

    def MAC_MAX_RETRY_NUMBER_INDIRECT(self, _, payload):
        return self.parse_C(payload)

    def NET_SAVED(self, _, payload):
        return self.parse_b(payload)

    def NET_IF_UP(self, _, payload):
        return self.parse_b(payload)

    def NET_STACK_UP(self, _, payload):
        return self.parse_C(payload)

    def NET_ROLE(self, _, payload):
        return self.parse_C(payload)

    def NET_NETWORK_NAME(self, _, payload):
        return self.parse_U(payload)

    def NET_XPANID(self, _, payload):
        return self.parse_D(payload)

    def NET_MASTER_KEY(self, _, payload):
        return self.parse_D(payload)

    def NET_KEY_SEQUENCE_COUNTER(self, _, payload):
        return self.parse_L(payload)

    def NET_PARTITION_ID(self, _, payload):
        return self.parse_L(payload)

    def NET_KEY_SWITCH_GUARDTIME(self, _, payload):
        return self.parse_L(payload)

    def THREAD_LEADER_ADDR(self, _, payload):
        return self.parse_6(payload)

    def THREAD_PARENT(self, _wpan_api, payload):
        return self.parse_fields(payload, "ES")

    def THREAD_CHILD_TABLE(self, _, payload):
        return self.parse_fields(payload, "A(t(ESLLCCcCc))")

    def THREAD_LEADER_RID(self, _, payload):
        return self.parse_C(payload)

    def THREAD_LEADER_WEIGHT(self, _, payload):
        return self.parse_C(payload)

    def THREAD_LOCAL_LEADER_WEIGHT(self, _, payload):
        return self.parse_C(payload)

    def THREAD_NETWORK_DATA(self, _, payload):
        return self.parse_D(payload)

    def THREAD_NETWORK_DATA_VERSION(self, _wpan_api, payload):
        return self.parse_C(payload)

    def THREAD_STABLE_NETWORK_DATA(self, _wpan_api, payload):
        pass

    def THREAD_STABLE_NETWORK_DATA_VERSION(self, _wpan_api, payload):
        return self.parse_C(payload)

    def __init__(self):
        self.autoAddresses = set()

        self.wpan_api = None
        self.__queue_prefix = queue.Queue()
        self.prefix_thread = threading.Thread(target=self.__run_prefix_handler)
        self.prefix_thread.setDaemon(True)
        self.prefix_thread.start()

    def handle_prefix_change(self, payload):
        """ Automatically ipaddr add / remove addresses for each new prefix. """
        # As done by cli.cpp Interpreter::HandleNetifStateChanged

        # First parse payload and extract slaac prefix information.
        pay = payload
        Prefix = namedtuple("Prefix", "prefix prefixlen stable flags is_local")
        prefixes = []
        slaacPrefixSet = set()
        while len(pay) >= 22:
            (_structlen) = unpack('<H', pay[:2])
            struct_len = _structlen[0]
            pay = pay[2:]
            prefix = Prefix(*unpack('16sBBBB', pay[:20]))
            if prefix.flags & kThread.PrefixSlaacFlag:
                net6 = ipaddress.IPv6Network(prefix.prefix)
                net6 = net6.supernet(new_prefix=prefix.prefixlen)
                slaacPrefixSet.add(net6)
                prefixes.append(prefix)
            pay = pay[struct_len:]

        if CONFIG.DEBUG_LOG_PROP:
            print("\n========= PREFIX ============")
            print("ipaddrs: " + str(self.autoAddresses))
            print("slaac prefix set: " + str(slaacPrefixSet))
            print("==============================\n")

    def __run_prefix_handler(self):
        while 1:
            (wpan_api, payload) = self.__queue_prefix.get(True)
            self.wpan_api = wpan_api
            self.handle_prefix_change(payload)
            self.__queue_prefix.task_done()

    def THREAD_ON_MESH_NETS(self, wpan_api, payload):
        if FEATURE_USE_SLACC:
            # Kick prefix handler thread to allow serial rx thread to work.
            self.__queue_prefix.put_nowait((wpan_api, payload))

        return self.parse_D(payload)

    def THREAD_LOCAL_ROUTES(self, _wpan_api, payload):
        pass

    def THREAD_ASSISTING_PORTS(self, _wpan_api, payload):
        pass

    def THREAD_ALLOW_LOCAL_NET_DATA_CHANGE(self, _, payload):
        return self.parse_b(payload)

    def THREAD_MODE(self, _, payload):
        return self.parse_C(payload)

    def THREAD_CHILD_COUNT_MAX(self, _, payload):
        return self.parse_C(payload)

    def THREAD_CHILD_TIMEOUT(self, _, payload):
        return self.parse_L(payload)

    def THREAD_RLOC16(self, _, payload):
        return self.parse_S(payload)

    def THREAD_ROUTER_UPGRADE_THRESHOLD(self, _, payload):
        return self.parse_C(payload)

    def THREAD_ROUTER_DOWNGRADE_THRESHOLD(self, _, payload):
        return self.parse_C(payload)

    def THREAD_ROUTER_SELECTION_JITTER(self, _, payload):
        return self.parse_C(payload)

    def THREAD_NEIGHBOR_TABLE(self, _, payload):
        return self.parse_fields(payload, 'A(t(ESLCcCbLL))')

    def THREAD_CONTEXT_REUSE_DELAY(self, _, payload):
        return self.parse_L(payload)

    def THREAD_NETWORK_ID_TIMEOUT(self, _, payload):
        return self.parse_C(payload)

    def THREAD_ACTIVE_ROUTER_IDS(self, _, payload):
        return self.parse_D(payload)

    def THREAD_RLOC16_DEBUG_PASSTHRU(self, _, payload):
        return self.parse_b(payload)

    def MESHCOP_JOINER_ENABLE(self, _, payload):
        return self.parse_b(payload)

    def MESHCOP_JOINER_CREDENTIAL(self, _, payload):
        return self.parse_D(payload)

    def MESHCOP_JOINER_URL(self, _, payload):
        return self.parse_U(payload)

    def MESHCOP_BORDER_AGENT_ENABLE(self, _, payload):
        return self.parse_b(payload)

    def IPV6_LL_ADDR(self, _, payload):
        return self.parse_6(payload)

    def IPV6_ML_ADDR(self, _, payload):
        return self.parse_6(payload)

    def IPV6_ML_PREFIX(self, _, payload):
        return self.parse_E(payload)

    def IPV6_ADDRESS_TABLE(self, _, payload):
        return self.parse_D(payload)

    def IPV6_ROUTE_TABLE(self, _, payload):
        return self.parse_D(payload)

    def IPv6_ICMP_PING_OFFLOAD(self, _, payload):
        return self.parse_b(payload)

    def STREAM_DEBUG(self, _, payload):
        return self.parse_D(payload)

    def STREAM_RAW(self, _, payload):
        return self.parse_D(payload)

    def STREAM_NET(self, _, payload):
        return self.parse_d(payload)

    def STREAM_NET_INSECURE(self, _, payload):
        return self.parse_d(payload)

    def STREAM_LOG(self, _, payload):
        return self.parse_fields(payload, "UD")

    def PIB_PHY_CHANNELS_SUPPORTED(self, _wpan_api, payload):
        pass

    def PIB_MAC_PROMISCUOUS_MODE(self, _wpan_api, payload):
        pass

    def PIB_MAC_SECURITY_ENABLED(self, _wpan_api, payload):
        pass

    def MSG_BUFFER_COUNTERS(self, _wpan_api, payload):
        return self.parse_fields(payload, "SSSSSSSSSSSSSSSS")

    def ALL_MAC_COUNTERS(self, _wpan_api, payload):
        return self.parse_fields(payload, "t(A(L))t(A(L))")

    def MLE_COUNTERS(self, _wpan_api, payload):
        return self.parse_fields(payload, "SSSSSSSSS")

    def MAC_RETRY_HISTOGRAM(self, _wpan_api, payload):
        return self.parse_fields(payload, "t(A(L))t(A(L))")

    def NEST_STREAM_MFG(self, _wpan_api, payload):
        return self.parse_U(payload)


#=========================================


class SpinelCommandHandler(SpinelCodec):

    def handle_prop(self, wpan_api, name, payload, tid):
        (prop_id, prop_len) = self.parse_i(payload)

        if prop_id in SPINEL_PROP_DISPATCH:
            handler = SPINEL_PROP_DISPATCH[prop_id]
            prop_name = handler.__name__

            # Skip any VALUE_INSERTED(CHILD_TABLE) or VALUE_REMOVED(CHILD_TABLE)
            if prop_id == SPINEL.PROP_THREAD_CHILD_TABLE:
                if name in ["INSERTED", "REMOVED"]:
                    return

            prop_value = handler(wpan_api, payload[prop_len:])

            if CONFIG.DEBUG_LOG_PROP:

                # Generic output
                if isinstance(prop_value, str):
                    prop_value_str = util.hexify_str(prop_value)
                    CONFIG.LOGGER.debug("PROP_VALUE_%s [tid=%d]: %s = %s", name,
                                        (tid & 0xF), prop_name, prop_value_str)
                else:
                    prop_value_str = str(prop_value)

                    CONFIG.LOGGER.debug("PROP_VALUE_%s [tid=%d]: %s = %s", name,
                                        (tid & 0xF), prop_name, prop_value_str)

                # Extend output for certain properties.
                if prop_id == SPINEL.PROP_LAST_STATUS:
                    CONFIG.LOGGER.debug(SPINEL_LAST_STATUS_MAP[prop_value])

            if CONFIG.DEBUG_LOG_PKT:
                if ((prop_id == SPINEL.PROP_STREAM_NET) or
                    (prop_id == SPINEL.PROP_STREAM_NET_INSECURE)):
                    CONFIG.LOGGER.debug("PROP_VALUE_" + name + ": " + prop_name)

                elif prop_id == SPINEL.PROP_STREAM_DEBUG:
                    CONFIG.LOGGER.debug("DEBUG: " + prop_value)

            if wpan_api:
                wpan_api.queue_add(prop_id, prop_value, tid)
            else:
                print("no wpan_api")
        elif CONFIG.DEBUG_LOG_PROP:
            prop_name = "Property Unknown"
            CONFIG.LOGGER.info("\n%s (%i): ", prop_name, prop_id)

    def PROP_VALUE_IS(self, wpan_api, payload, tid):
        self.handle_prop(wpan_api, "IS", payload, tid)

    def PROP_VALUE_INSERTED(self, wpan_api, payload, tid):
        self.handle_prop(wpan_api, "INSERTED", payload, tid)

    def PROP_VALUE_REMOVED(self, wpan_api, payload, tid):
        self.handle_prop(wpan_api, "REMOVED", payload, tid)


WPAN_CMD_HANDLER = SpinelCommandHandler()

SPINEL_COMMAND_DISPATCH = {
    SPINEL.RSP_PROP_VALUE_IS: WPAN_CMD_HANDLER.PROP_VALUE_IS,
    SPINEL.RSP_PROP_VALUE_INSERTED: WPAN_CMD_HANDLER.PROP_VALUE_INSERTED,
    SPINEL.RSP_PROP_VALUE_REMOVED: WPAN_CMD_HANDLER.PROP_VALUE_REMOVED,
}

WPAN_PROP_HANDLER = SpinelPropertyHandler()

SPINEL_PROP_DISPATCH = {
    SPINEL.PROP_LAST_STATUS:
        WPAN_PROP_HANDLER.LAST_STATUS,
    SPINEL.PROP_PROTOCOL_VERSION:
        WPAN_PROP_HANDLER.PROTOCOL_VERSION,
    SPINEL.PROP_NCP_VERSION:
        WPAN_PROP_HANDLER.NCP_VERSION,
    SPINEL.PROP_INTERFACE_TYPE:
        WPAN_PROP_HANDLER.INTERFACE_TYPE,
    SPINEL.PROP_VENDOR_ID:
        WPAN_PROP_HANDLER.VENDOR_ID,
    SPINEL.PROP_CAPS:
        WPAN_PROP_HANDLER.CAPS,
    SPINEL.PROP_INTERFACE_COUNT:
        WPAN_PROP_HANDLER.INTERFACE_COUNT,
    SPINEL.PROP_POWER_STATE:
        WPAN_PROP_HANDLER.POWER_STATE,
    SPINEL.PROP_HWADDR:
        WPAN_PROP_HANDLER.HWADDR,
    SPINEL.PROP_LOCK:
        WPAN_PROP_HANDLER.LOCK,
    SPINEL.PROP_HBO_MEM_MAX:
        WPAN_PROP_HANDLER.HBO_MEM_MAX,
    SPINEL.PROP_HBO_BLOCK_MAX:
        WPAN_PROP_HANDLER.HBO_BLOCK_MAX,
    SPINEL.PROP_PHY_ENABLED:
        WPAN_PROP_HANDLER.PHY_ENABLED,
    SPINEL.PROP_PHY_CHAN:
        WPAN_PROP_HANDLER.PHY_CHAN,
    SPINEL.PROP_PHY_CHAN_SUPPORTED:
        WPAN_PROP_HANDLER.PHY_CHAN_SUPPORTED,
    SPINEL.PROP_PHY_FREQ:
        WPAN_PROP_HANDLER.PHY_FREQ,
    SPINEL.PROP_PHY_CCA_THRESHOLD:
        WPAN_PROP_HANDLER.PHY_CCA_THRESHOLD,
    SPINEL.PROP_PHY_TX_POWER:
        WPAN_PROP_HANDLER.PHY_TX_POWER,
    SPINEL.PROP_PHY_RSSI:
        WPAN_PROP_HANDLER.PHY_RSSI,
    SPINEL.PROP_MAC_SCAN_STATE:
        WPAN_PROP_HANDLER.MAC_SCAN_STATE,
    SPINEL.PROP_MAC_SCAN_MASK:
        WPAN_PROP_HANDLER.MAC_SCAN_MASK,
    SPINEL.PROP_MAC_SCAN_PERIOD:
        WPAN_PROP_HANDLER.MAC_SCAN_PERIOD,
    SPINEL.PROP_MAC_SCAN_BEACON:
        WPAN_PROP_HANDLER.MAC_SCAN_BEACON,
    SPINEL.PROP_MAC_15_4_LADDR:
        WPAN_PROP_HANDLER.MAC_15_4_LADDR,
    SPINEL.PROP_MAC_15_4_SADDR:
        WPAN_PROP_HANDLER.MAC_15_4_SADDR,
    SPINEL.PROP_MAC_15_4_PANID:
        WPAN_PROP_HANDLER.MAC_15_4_PANID,
    SPINEL.PROP_MAC_RAW_STREAM_ENABLED:
        WPAN_PROP_HANDLER.MAC_RAW_STREAM_ENABLED,
    SPINEL.PROP_MAC_FILTER_MODE:
        WPAN_PROP_HANDLER.MAC_FILTER_MODE,
    SPINEL.PROP_MAC_ALLOWLIST:
        WPAN_PROP_HANDLER.MAC_ALLOWLIST,
    SPINEL.PROP_MAC_ALLOWLIST_ENABLED:
        WPAN_PROP_HANDLER.MAC_ALLOWLIST_ENABLED,
    SPINEL.PROP_MAC_DENYLIST:
        WPAN_PROP_HANDLER.MAC_DENYLIST,
    SPINEL.PROP_MAC_DENYLIST_ENABLED:
        WPAN_PROP_HANDLER.MAC_DENYLIST_ENABLED,
    SPINEL.PROP_MAC_FIXED_RSS:
        WPAN_PROP_HANDLER.MAC_FIXED_RSS,
    SPINEL.PROP_MAC_MAX_RETRY_NUMBER_DIRECT:
        WPAN_PROP_HANDLER.MAC_MAX_RETRY_NUMBER_DIRECT,
    SPINEL.PROP_MAC_MAX_RETRY_NUMBER_INDIRECT:
        WPAN_PROP_HANDLER.MAC_MAX_RETRY_NUMBER_INDIRECT,
    SPINEL.PROP_NET_SAVED:
        WPAN_PROP_HANDLER.NET_SAVED,
    SPINEL.PROP_NET_IF_UP:
        WPAN_PROP_HANDLER.NET_IF_UP,
    SPINEL.PROP_NET_STACK_UP:
        WPAN_PROP_HANDLER.NET_STACK_UP,
    SPINEL.PROP_NET_ROLE:
        WPAN_PROP_HANDLER.NET_ROLE,
    SPINEL.PROP_NET_NETWORK_NAME:
        WPAN_PROP_HANDLER.NET_NETWORK_NAME,
    SPINEL.PROP_NET_XPANID:
        WPAN_PROP_HANDLER.NET_XPANID,
    SPINEL.PROP_NET_MASTER_KEY:
        WPAN_PROP_HANDLER.NET_MASTER_KEY,
    SPINEL.PROP_NET_KEY_SEQUENCE_COUNTER:
        WPAN_PROP_HANDLER.NET_KEY_SEQUENCE_COUNTER,
    SPINEL.PROP_NET_PARTITION_ID:
        WPAN_PROP_HANDLER.NET_PARTITION_ID,
    SPINEL.PROP_NET_KEY_SWITCH_GUARDTIME:
        WPAN_PROP_HANDLER.NET_KEY_SWITCH_GUARDTIME,
    SPINEL.PROP_THREAD_LEADER_ADDR:
        WPAN_PROP_HANDLER.THREAD_LEADER_ADDR,
    SPINEL.PROP_THREAD_PARENT:
        WPAN_PROP_HANDLER.THREAD_PARENT,
    SPINEL.PROP_THREAD_CHILD_TABLE:
        WPAN_PROP_HANDLER.THREAD_CHILD_TABLE,
    SPINEL.PROP_THREAD_LEADER_RID:
        WPAN_PROP_HANDLER.THREAD_LEADER_RID,
    SPINEL.PROP_THREAD_LEADER_WEIGHT:
        WPAN_PROP_HANDLER.THREAD_LEADER_WEIGHT,
    SPINEL.PROP_THREAD_LOCAL_LEADER_WEIGHT:
        WPAN_PROP_HANDLER.THREAD_LOCAL_LEADER_WEIGHT,
    SPINEL.PROP_THREAD_NETWORK_DATA:
        WPAN_PROP_HANDLER.THREAD_NETWORK_DATA,
    SPINEL.PROP_THREAD_NETWORK_DATA_VERSION:
        WPAN_PROP_HANDLER.THREAD_NETWORK_DATA_VERSION,
    SPINEL.PROP_THREAD_STABLE_NETWORK_DATA:
        WPAN_PROP_HANDLER.THREAD_STABLE_NETWORK_DATA,
    SPINEL.PROP_THREAD_STABLE_NETWORK_DATA_VERSION:
        WPAN_PROP_HANDLER.THREAD_STABLE_NETWORK_DATA_VERSION,
    SPINEL.PROP_THREAD_ON_MESH_NETS:
        WPAN_PROP_HANDLER.THREAD_ON_MESH_NETS,
    SPINEL.PROP_THREAD_LOCAL_ROUTES:
        WPAN_PROP_HANDLER.THREAD_LOCAL_ROUTES,
    SPINEL.PROP_THREAD_ASSISTING_PORTS:
        WPAN_PROP_HANDLER.THREAD_ASSISTING_PORTS,
    SPINEL.PROP_THREAD_ALLOW_LOCAL_NET_DATA_CHANGE:
        WPAN_PROP_HANDLER.THREAD_ALLOW_LOCAL_NET_DATA_CHANGE,
    SPINEL.PROP_THREAD_MODE:
        WPAN_PROP_HANDLER.THREAD_MODE,
    SPINEL.PROP_THREAD_CHILD_COUNT_MAX:
        WPAN_PROP_HANDLER.THREAD_CHILD_COUNT_MAX,
    SPINEL.PROP_THREAD_CHILD_TIMEOUT:
        WPAN_PROP_HANDLER.THREAD_CHILD_TIMEOUT,
    SPINEL.PROP_THREAD_RLOC16:
        WPAN_PROP_HANDLER.THREAD_RLOC16,
    SPINEL.PROP_THREAD_ROUTER_UPGRADE_THRESHOLD:
        WPAN_PROP_HANDLER.THREAD_ROUTER_UPGRADE_THRESHOLD,
    SPINEL.PROP_THREAD_ROUTER_DOWNGRADE_THRESHOLD:
        WPAN_PROP_HANDLER.THREAD_ROUTER_DOWNGRADE_THRESHOLD,
    SPINEL.PROP_THREAD_ROUTER_SELECTION_JITTER:
        WPAN_PROP_HANDLER.THREAD_ROUTER_SELECTION_JITTER,
    SPINEL.PROP_THREAD_CONTEXT_REUSE_DELAY:
        WPAN_PROP_HANDLER.THREAD_CONTEXT_REUSE_DELAY,
    SPINEL.PROP_THREAD_NETWORK_ID_TIMEOUT:
        WPAN_PROP_HANDLER.THREAD_NETWORK_ID_TIMEOUT,
    SPINEL.PROP_THREAD_ACTIVE_ROUTER_IDS:
        WPAN_PROP_HANDLER.THREAD_ACTIVE_ROUTER_IDS,
    SPINEL.PROP_THREAD_RLOC16_DEBUG_PASSTHRU:
        WPAN_PROP_HANDLER.THREAD_RLOC16_DEBUG_PASSTHRU,
    SPINEL.PROP_THREAD_NEIGHBOR_TABLE:
        WPAN_PROP_HANDLER.THREAD_NEIGHBOR_TABLE,
    SPINEL.PROP_MESHCOP_JOINER_ENABLE:
        WPAN_PROP_HANDLER.MESHCOP_JOINER_ENABLE,
    SPINEL.PROP_MESHCOP_JOINER_CREDENTIAL:
        WPAN_PROP_HANDLER.MESHCOP_JOINER_CREDENTIAL,
    SPINEL.PROP_MESHCOP_JOINER_URL:
        WPAN_PROP_HANDLER.MESHCOP_JOINER_URL,
    SPINEL.PROP_MESHCOP_BORDER_AGENT_ENABLE:
        WPAN_PROP_HANDLER.MESHCOP_BORDER_AGENT_ENABLE,
    SPINEL.PROP_IPV6_LL_ADDR:
        WPAN_PROP_HANDLER.IPV6_LL_ADDR,
    SPINEL.PROP_IPV6_ML_ADDR:
        WPAN_PROP_HANDLER.IPV6_ML_ADDR,
    SPINEL.PROP_IPV6_ML_PREFIX:
        WPAN_PROP_HANDLER.IPV6_ML_PREFIX,
    SPINEL.PROP_IPV6_ADDRESS_TABLE:
        WPAN_PROP_HANDLER.IPV6_ADDRESS_TABLE,
    SPINEL.PROP_IPV6_ROUTE_TABLE:
        WPAN_PROP_HANDLER.IPV6_ROUTE_TABLE,
    SPINEL.PROP_IPv6_ICMP_PING_OFFLOAD:
        WPAN_PROP_HANDLER.IPv6_ICMP_PING_OFFLOAD,
    SPINEL.PROP_STREAM_DEBUG:
        WPAN_PROP_HANDLER.STREAM_DEBUG,
    SPINEL.PROP_STREAM_RAW:
        WPAN_PROP_HANDLER.STREAM_RAW,
    SPINEL.PROP_STREAM_NET:
        WPAN_PROP_HANDLER.STREAM_NET,
    SPINEL.PROP_STREAM_NET_INSECURE:
        WPAN_PROP_HANDLER.STREAM_NET_INSECURE,
    SPINEL.PROP_STREAM_LOG:
        WPAN_PROP_HANDLER.STREAM_LOG,
    SPINEL.PROP_PIB_15_4_PHY_CHANNELS_SUPPORTED:
        WPAN_PROP_HANDLER.PIB_PHY_CHANNELS_SUPPORTED,
    SPINEL.PROP_PIB_15_4_MAC_PROMISCUOUS_MODE:
        WPAN_PROP_HANDLER.PIB_MAC_PROMISCUOUS_MODE,
    SPINEL.PROP_PIB_15_4_MAC_SECURITY_ENABLED:
        WPAN_PROP_HANDLER.PIB_MAC_SECURITY_ENABLED,
    SPINEL.PROP_MSG_BUFFER_COUNTERS:
        WPAN_PROP_HANDLER.MSG_BUFFER_COUNTERS,
    SPINEL.PROP_CNTR_ALL_MAC_COUNTERS:
        WPAN_PROP_HANDLER.ALL_MAC_COUNTERS,
    SPINEL.PROP_CNTR_MLE_COUNTERS:
        WPAN_PROP_HANDLER.MLE_COUNTERS,
    SPINEL.PROP_CNTR_MAC_RETRY_HISTOGRAM:
        WPAN_PROP_HANDLER.MAC_RETRY_HISTOGRAM,
    SPINEL.PROP_NEST_STREAM_MFG:
        WPAN_PROP_HANDLER.NEST_STREAM_MFG
}


class WpanApi(SpinelCodec):
    """ Helper class to format wpan command packets """

    def __init__(self,
                 stream,
                 nodeid,
                 use_hdlc=FEATURE_USE_HDLC,
                 timeout=TIMEOUT_PROP,
                 vendor_module=None):
        self.stream = stream
        self.nodeid = nodeid

        self.timeout = timeout

        self.use_hdlc = use_hdlc
        if self.use_hdlc:
            self.hdlc = Hdlc(self.stream)

        if vendor_module:
            # Hook vendor properties
            try:
                codec = importlib.import_module(vendor_module + '.codec')
                SPINEL_PROP_DISPATCH.update(codec.VENDOR_SPINEL_PROP_DISPATCH)
            except ImportError:
                pass

        # PARSER state
        self.rx_pkt = []
        self.callback = defaultdict(list)  # Map prop_id to list of callbacks.

        # Fire up threads
        self._reader_alive = True
        self.tid_filter = set()
        self.__queue_prop = defaultdict(queue.Queue)  # Map tid to Queue.
        self.queue_register()
        self.__start_reader()

    def __del__(self):
        self._reader_alive = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._reader_alive = False

    def __start_reader(self):
        """Start reader thread"""
        self._reader_alive = True
        # start serial->console thread
        self.receiver_thread = threading.Thread(target=self.stream_rx)
        self.receiver_thread.setDaemon(True)
        self.receiver_thread.start()

    def transact(self, command_id, payload=bytes(), tid=SPINEL.HEADER_DEFAULT):
        pkt = self.encode_packet(command_id, payload, tid)
        if CONFIG.DEBUG_LOG_SERIAL:
            msg = "TX Pay: (%i) %s " % (len(pkt),
                                        binascii.hexlify(pkt).decode('utf-8'))
            CONFIG.LOGGER.debug(msg)

        if self.use_hdlc:
            pkt = self.hdlc.encode(pkt)
        self.stream_tx(pkt)

    def parse_rx(self, pkt):
        if not pkt:
            return

        if CONFIG.DEBUG_LOG_SERIAL:
            msg = "RX Pay: (%i) %s " % (len(pkt),
                                        binascii.hexlify(pkt).decode('utf-8'))
            CONFIG.LOGGER.debug(msg)

        length = len(pkt) - 2
        if length < 0:
            return

        spkt = pkt

        #if not isinstance(spkt, str):
        #    spkt = "".join(map(chr, spkt))

        tid = self.parse_C(spkt[:1])
        (cmd_id, cmd_length) = self.parse_i(spkt[1:])
        pay_start = cmd_length + 1
        payload = spkt[pay_start:]

        try:
            handler = SPINEL_COMMAND_DISPATCH[cmd_id]
            cmd_name = handler.__name__
            handler(self, payload, tid)

        except Exception as _ex:
            print(traceback.format_exc())
            cmd_name = "CB_Unknown"
            CONFIG.LOGGER.info("\n%s (%i): ", cmd_name, cmd_id)

        if CONFIG.DEBUG_CMD_RESPONSE:
            CONFIG.LOGGER.info("\n%s (%i): ", cmd_name, cmd_id)
            CONFIG.LOGGER.info("===> %s",
                               binascii.hexlify(payload).decode('utf-8'))

    def stream_tx(self, pkt):
        # Encapsulate lagging and Framer support in self.stream class.
        self.stream.write(pkt)

    def stream_rx(self):
        """ Recieve thread and parser. """
        try:
            while self._reader_alive:
                if self.use_hdlc:
                    self.rx_pkt = self.hdlc.collect()
                else:
                    # size=None: Assume stream will always deliver packets
                    pkt = self.stream.read(None)
                    self.rx_pkt = util.packed_to_array(pkt)

                self.parse_rx(self.rx_pkt)
        except:
            if self._reader_alive:
                raise
            else:
                # Ignore the error since we are exiting
                pass

    class PropertyItem(object):
        """ Queue item for NCP response to property commands. """

        def __init__(self, prop, value, tid):
            self.prop = prop
            self.value = value
            self.tid = tid

    def callback_register(self, prop, cb):
        self.callback[prop].append(cb)

    def queue_register(self, tid=SPINEL.HEADER_DEFAULT):
        self.tid_filter.add(tid)
        return self.__queue_prop[tid]

    def queue_wait_prepare(self, _prop_id, tid=SPINEL.HEADER_DEFAULT):
        self.queue_clear(tid)

    def queue_add(self, prop, value, tid):
        cb_list = self.callback[prop]

        # Asynchronous handlers can consume message and not add to queue.
        if len(cb_list) > 0:
            consumed = cb_list[0](prop, value, tid)
            if consumed:
                return

        if tid not in self.tid_filter:
            return
        item = self.PropertyItem(prop, value, tid)
        self.__queue_prop[tid].put_nowait(item)

    def queue_clear(self, tid):
        with self.__queue_prop[tid].mutex:
            self.__queue_prop[tid].queue.clear()

    def queue_get(self, tid, timeout=None):
        try:
            if (timeout):
                item = self.__queue_prop[tid].get(True, timeout)
            else:
                item = self.__queue_prop[tid].get_nowait()
        except queue.Empty:
            item = None
        return item

    def queue_wait_for_prop(self,
                            _prop,
                            tid=SPINEL.HEADER_DEFAULT,
                            timeout=None):
        if _prop is None:
            return None

        if timeout is None:
            timeout = self.timeout

        processed_queue = queue.Queue()
        timeout_time = time.time() + timeout

        while time.time() < timeout_time:
            item = self.queue_get(tid, timeout_time - time.time())

            if item is None:
                continue
            if item.prop == _prop:
                break

            processed_queue.put_nowait(item)
        else:
            item = None

        # To make sure that all received properties will be processed in the same order.
        with self.__queue_prop[tid].mutex:
            while self.__queue_prop[tid]._qsize() > 0:
                processed_queue.put(self.__queue_prop[tid]._get())

            while not processed_queue.empty():
                self.__queue_prop[tid]._put(processed_queue.get_nowait())

        return item

    def ip_send(self, pkt):
        pay = self.encode_i(SPINEL.PROP_STREAM_NET)

        pkt_len = len(pkt)
        pay += pack("<H", pkt_len)  # Start with length of IPv6 packet

        pkt_len += 2  # Increment to include length word
        pay += pkt  # Append packet after length

        self.transact(SPINEL.CMD_PROP_VALUE_SET, pay)

    def cmd_reset(self):
        self.queue_wait_prepare(None, SPINEL.HEADER_ASYNC)
        self.transact(SPINEL.CMD_RESET)
        result = self.queue_wait_for_prop(SPINEL.PROP_LAST_STATUS,
                                          SPINEL.HEADER_ASYNC)
        return (result is not None and result.value == 114)

    def cmd_send(self, command_id, payload=bytes(), tid=SPINEL.HEADER_DEFAULT):
        self.queue_wait_prepare(None, tid)
        self.transact(command_id, payload, tid)
        self.queue_wait_for_prop(None, tid)

    def prop_change_async(self,
                          cmd,
                          prop_id,
                          value,
                          py_format='B',
                          tid=SPINEL.HEADER_DEFAULT):
        pay = self.encode_i(prop_id)
        if py_format != None:
            pay += pack(py_format, value)
        self.transact(cmd, pay, tid)

    def prop_insert_async(self,
                          prop_id,
                          value,
                          py_format='B',
                          tid=SPINEL.HEADER_DEFAULT):
        self.prop_change_async(SPINEL.CMD_PROP_VALUE_INSERT, prop_id, value,
                               py_format, tid)

    def prop_remove_async(self,
                          prop_id,
                          value,
                          py_format='B',
                          tid=SPINEL.HEADER_DEFAULT):
        self.prop_change_async(SPINEL.CMD_PROP_VALUE_REMOVE, prop_id, value,
                               py_format, tid)

    def __prop_change_value(self,
                            cmd,
                            prop_id,
                            value,
                            py_format='B',
                            tid=SPINEL.HEADER_DEFAULT):
        """ Utility routine to change a property value over SPINEL. """
        self.queue_wait_prepare(prop_id, tid)

        pay = self.encode_i(prop_id)
        if py_format != None:
            pay += pack(py_format, value)
        self.transact(cmd, pay, tid)

        result = self.queue_wait_for_prop(prop_id, tid)
        if result:
            return result.value
        else:
            return None

    def prop_get_value(self, prop_id, tid=SPINEL.HEADER_DEFAULT):
        """ Blocking routine to get a property value over SPINEL. """
        if CONFIG.DEBUG_LOG_PROP:
            handler = SPINEL_PROP_DISPATCH[prop_id]
            prop_name = handler.__name__
            print("PROP_VALUE_GET [tid=%d]: %s" % (tid & 0xF, prop_name))
        return self.__prop_change_value(SPINEL.CMD_PROP_VALUE_GET, prop_id,
                                        None, None, tid)

    def prop_set_value(self,
                       prop_id,
                       value,
                       py_format='B',
                       tid=SPINEL.HEADER_DEFAULT):
        """ Blocking routine to set a property value over SPINEL. """
        if CONFIG.DEBUG_LOG_PROP:
            handler = SPINEL_PROP_DISPATCH[prop_id]
            prop_name = handler.__name__
            print("PROP_VALUE_SET [tid=%d]: %s" % (tid & 0xF, prop_name))
        return self.__prop_change_value(SPINEL.CMD_PROP_VALUE_SET, prop_id,
                                        value, py_format, tid)

    def prop_insert_value(self,
                          prop_id,
                          value,
                          py_format='B',
                          tid=SPINEL.HEADER_DEFAULT):
        """ Blocking routine to insert a property value over SPINEL. """
        if CONFIG.DEBUG_LOG_PROP:
            handler = SPINEL_PROP_DISPATCH[prop_id]
            prop_name = handler.__name__
            print("PROP_VALUE_INSERT [tid=%d]: %s" % (tid & 0xF, prop_name))
        return self.__prop_change_value(SPINEL.CMD_PROP_VALUE_INSERT, prop_id,
                                        value, py_format, tid)

    def prop_remove_value(self,
                          prop_id,
                          value,
                          py_format='B',
                          tid=SPINEL.HEADER_DEFAULT):
        """ Blocking routine to remove a property value over SPINEL. """
        if CONFIG.DEBUG_LOG_PROP:
            handler = SPINEL_PROP_DISPATCH[prop_id]
            prop_name = handler.__name__
            print("PROP_VALUE_REMOVE [tid=%d]: %s" % (tid & 0xF, prop_name))
        return self.__prop_change_value(SPINEL.CMD_PROP_VALUE_REMOVE, prop_id,
                                        value, py_format, tid)

    def get_ipaddrs(self, tid=SPINEL.HEADER_DEFAULT):
        """
        Return current list of ip addresses for the device.
        """
        value = self.prop_get_value(SPINEL.PROP_IPV6_ADDRESS_TABLE, tid)
        # TODO: clean up table parsing to be less hard-coded magic.
        if value is None:
            return None
        size = 0x1B
        addrs = [value[i:i + size] for i in range(0, len(value), size)]
        ipaddrs = []
        for addr in addrs:
            addr = addr[2:18]
            ipaddrs.append(ipaddress.IPv6Address(addr))
        return ipaddrs



#!/usr/bin/env python
#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import binascii


def hexify_str(s, delim=':'):
    hex_str = binascii.hexlify(s.encode('utf-8')).decode('utf-8')
    return delim.join([hex_str[i:i + 2] for i in range(0, len(hex_str), 2)])


def packed_to_array(packet):
    return list(map(ord, packet))


def asciify_int(i):
    return "%c" % (i)


def hex_to_bytes(s):
    result = bytes()

    for i in range(0, len(s), 2):
        (b1, b2) = s[i:i + 2]
        hex = b1 + b2
        v = int(hex, 16)
        result += bytearray([v])

    return result



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
""" High-Level Data Link Control (HDLC) module. """

import binascii

from struct import pack

import spinel.config as CONFIG
from spinel.stream import IStream

HDLC_FLAG = 0x7e
HDLC_ESCAPE = 0x7d

# RFC 1662 Appendix C

HDLC_FCS_INIT = 0xFFFF
HDLC_FCS_POLY = 0x8408
HDLC_FCS_GOOD = 0xF0B8


class Hdlc(IStream):
    """ Utility class for HDLC encoding and decoding. """

    def __init__(self, stream):
        self.stream = stream
        self.fcstab = self.mkfcstab()

    @classmethod
    def mkfcstab(cls):
        """ Make a static lookup table for byte value to FCS16 result. """
        polynomial = HDLC_FCS_POLY

        def valiter():
            """ Helper to yield FCS16 table entries for each byte value. """
            for byte in range(256):
                fcs = byte
                i = 8
                while i:
                    fcs = (fcs >> 1) ^ polynomial if fcs & 1 else fcs >> 1
                    i -= 1

                yield fcs & 0xFFFF

        return tuple(valiter())

    def fcs16(self, byte, fcs):
        """
        Return the next iteration of an fcs16 calculation
        given the next data byte and current fcs accumulator.
        """
        fcs = (fcs >> 8) ^ self.fcstab[(fcs ^ byte) & 0xff]
        return fcs

    def collect(self):
        """ Return the next valid packet to pass HDLC decoding on the stream. """
        fcs = HDLC_FCS_INIT
        packet = []
        raw = []

        # Synchronize
        while 1:
            byte = self.stream.read()
            if CONFIG.DEBUG_HDLC:
                raw.append(byte)
            if byte == HDLC_FLAG:
                break

        # Read packet, updating fcs, and escaping bytes as needed
        while 1:
            byte = self.stream.read()
            if CONFIG.DEBUG_HDLC:
                raw.append(byte)
            if byte == HDLC_FLAG:
                if len(packet) != 0:
                    break
                else:
                    # If multiple FLAG bytes in a row, keep looking for data.
                    continue
            if byte == HDLC_ESCAPE:
                byte = self.stream.read()
                if CONFIG.DEBUG_HDLC:
                    raw.append(byte)
                byte ^= 0x20
            packet.append(byte)
            fcs = self.fcs16(byte, fcs)

        if CONFIG.DEBUG_HDLC:
            CONFIG.LOGGER.debug(
                "RX Hdlc: " + binascii.hexlify(bytearray(raw)).decode('utf-8'))

        if fcs != HDLC_FCS_GOOD:
            packet = None
        else:
            packet = packet[:-2]  # remove FCS16 from end
            packet = pack("%dB" % len(packet), *packet)

        return packet

    @classmethod
    def encode_byte(cls, byte, packet=[]):
        """ HDLC encode and append a single byte to the given packet. """
        if (byte == HDLC_ESCAPE) or (byte == HDLC_FLAG):
            packet.append(HDLC_ESCAPE)
            packet.append(byte ^ 0x20)
        else:
            packet.append(byte)
        return packet

    def encode(self, payload=""):
        """ Return the HDLC encoding of the given packet. """
        fcs = HDLC_FCS_INIT
        packet = []
        packet.append(HDLC_FLAG)
        for byte in payload:
            fcs = self.fcs16(byte, fcs)
            packet = self.encode_byte(byte, packet)

        fcs ^= 0xffff
        byte = fcs & 0xFF
        packet = self.encode_byte(byte, packet)
        byte = fcs >> 8
        packet = self.encode_byte(byte, packet)
        packet.append(HDLC_FLAG)
        packet = pack("%dB" % len(packet), *packet)

        if CONFIG.DEBUG_HDLC:
            CONFIG.LOGGER.debug(
                "TX Hdlc: " +
                binascii.hexlify(bytearray(packet)).decode('utf-8'))
        return packet

    def write(self, data):
        """ HDLC encode and write the given data to this stream. """
        pkt = self.encode(data)
        self.stream.write(pkt)

    def read(self, _size=None):
        """ Read and HDLC decode the next packet from this stream. """
        pkt = self.collect()
        return pkt



#
#  Copyright (c) 2016-2017, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
""" Unittest for spinel.codec module. """

import unittest

import spinel.util as util
from spinel.const import SPINEL
from spinel.codec import WpanApi
from spinel.test_stream import MockStream


class TestSniffer(unittest.TestCase):
    """ Unit TestCase class for sniffer relevant portions of spinel.codec.SpinelCodec. """

    HEADER = "800671"  # CMD_PROP_IS RAW_STREAM
    VECTOR = [
        # Some raw 6lo packets: ICMPv6EchoRequest to ff02::1, fe80::1, and MLE Advertisement
        "2d00499880fffffffffeff0d0100000001a7acdf3be9272c2d88765ff76f0bf08a7c3df0a78e9c1b23eb019c58740300800000",
        "3200699c81ffff0100000000000002feff0d030000000198e80cac00f8e0754e7542f5cb1171069f5c9689ef8d1d45a75e26b3f600800000",
        "450041d8980100ffffa8cb25ab2c32a0227f3b01f04d4c4d4cdc3b0015060000000000000001f226cce17968521d92904fec1adb0b94777030b944df65450bc955f05737e3901700800000"
    ]

    def test_prop_get(self):
        """ Unit test of SpinelCodec.prop_get_value. """

        mock_stream = MockStream({})

        nodeid = 1
        use_hdlc = False
        tid = SPINEL.HEADER_ASYNC
        prop_id = SPINEL.PROP_STREAM_RAW

        wpan_api = WpanApi(mock_stream, nodeid, use_hdlc)
        wpan_api.queue_register(tid)

        for truth in self.VECTOR:
            mock_stream.write_child_hex(self.HEADER + truth)
            result = wpan_api.queue_wait_for_prop(prop_id, tid)
            packet = util.hexify_str(result.value, "")
            self.failUnless(packet == truth)



#!/usr/bin/python
#
#  Copyright (c) 2019, The OpenThread Authors.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import os
import sys
import tempfile
import argparse
import subprocess
import threading
import logging
import re

from spinel.stream import StreamOpen
from spinel.const import SPINEL
from spinel.codec import WpanApi
from serial.tools.list_ports import comports
from enum import Enum

# Nodeid is required to execute ot-ncp-ftd for its sim radio socket port.
# This is maximum that works for MacOS.
DEFAULT_NODEID = 34
COMMON_BAUDRATE = [460800, 115200, 9600]


class Config(Enum):
    CHANNEL = 0
    BAUDRATE = 1
    TAP = 2


class _StreamCloser:

    def __init__(self, stream):
        self._stream = stream

    def __enter__(self):
        return self._stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stream.close()


def extcap_config(interface, option, extcap_version):
    """List Configuration for the given interface"""
    args = []
    values = []
    args.append(
        (Config.CHANNEL.value, '--channel', 'Channel', 'IEEE 802.15.4 channel',
         'selector', '{required=true}{default=11}'))

    match = re.match(r'^(\d+)(\.\d+)*$', extcap_version)
    if match and int(match.group(1)) >= 3:
        args.append((Config.TAP.value, '--tap',
                     'IEEE 802.15.4 TAP (only for Wireshark3.0 and later)',
                     'IEEE 802.15.4 TAP', 'boolflag', '{default=yes}'))

    for arg in args:
        print('arg {number=%d}{call=%s}{display=%s}{tooltip=%s}{type=%s}%s' %
              arg)

    values = values + [(Config.CHANNEL.value, '%d' % i, '%d' % i,
                        'true' if i == 11 else 'false') for i in range(11, 27)]

    for value in values:
        print('value {arg=%d}{value=%s}{display=%s}{default=%s}' % value)


def extcap_dlts(interface):
    """List DLTs for the given interface"""
    print(
        'dlt {number=195}{name=IEEE802_15_4_WITHFCS}{display=IEEE 802.15.4 with FCS}'
    )
    print('dlt {number=283}{name=IEEE802_15_4_TAP}{display=IEEE 802.15.4 TAP}')


def serialopen(interface, log_file):
    """
    Open serial to indentify OpenThread sniffer
    :param interface: string, eg: '/dev/ttyUSB0 - Zolertia Firefly platform', '/dev/ttyACM1 - nRF52840 OpenThread Device'
    """
    sys.stdout = log_file
    sys.stderr = log_file
    interface = str(interface).split()[0]
    baudrate = None

    for speed in COMMON_BAUDRATE:
        with _StreamCloser(StreamOpen('u', interface, False, baudrate=speed)) as stream, \
                WpanApi(stream, nodeid=DEFAULT_NODEID, timeout=0.1) as wpan_api:

            # result should not be None for both NCP and RCP
            result = wpan_api.prop_get_value(
                SPINEL.PROP_CAPS)  # confirm OpenThread Sniffer

            # check whether or not is OpenThread Sniffer
            if result is not None:
                baudrate = speed
                break

    if baudrate is not None:
        if sys.platform == 'win32':
            # Wireshark only shows the value of key `display`('OpenThread Sniffer').
            # Here intentionally appends interface in the end (e.g. 'OpenThread Sniffer: COM0').
            print('interface {value=%s:%s}{display=OpenThread Sniffer %s}' %
                  (interface, baudrate, interface),
                  file=sys.__stdout__,
                  flush=True)
        else:
            # On Linux or MacOS, wireshark will show the concatenation of the content of `display`
            # and `interface` by default (e.g. 'OpenThread Sniffer: /dev/ttyACM0').
            print('interface {value=%s:%s}{display=OpenThread Sniffer}' %
                  (interface, baudrate),
                  file=sys.__stdout__,
                  flush=True)


def extcap_interfaces():
    """List available interfaces to capture from"""

    log_file = open(
        os.path.join(tempfile.gettempdir(), 'extcap_ot_interfaces.log'), 'w')
    print(
        'extcap {version=1.0.0}{display=OpenThread Sniffer}{help=https://github.com/openthread/pyspinel}'
    )

    threads = []
    for interface in comports():
        th = threading.Thread(target=serialopen, args=(interface, log_file))
        threads.append(th)
        th.start()
    for th in threads:
        th.join()


def extcap_capture(interface, fifo, control_in, control_out, channel, tap):
    """Start the sniffer to capture packets"""
    # baudrate = detect_baudrate(interface)
    interface_port = str(interface).split(':')[0]
    interface_baudrate = str(interface).split(':')[1]

    with _StreamCloser(StreamOpen('u', interface_port, False, baudrate=int(interface_baudrate))) as stream, \
            WpanApi(stream, nodeid=DEFAULT_NODEID) as wpan_api:
        wpan_api.prop_set_value(SPINEL.PROP_PHY_ENABLED, 1)

    if sys.platform == 'win32':
        python_path = subprocess.Popen(
            'py -3 -c "import sys; print(sys.executable)"',
            stdout=subprocess.PIPE,
            shell=True,
        ).stdout.readline().decode().strip()
        sniffer_py = os.path.join(os.path.dirname(python_path), 'Scripts',
                                  'sniffer.py')
        cmd = ['python', sniffer_py]
    else:
        cmd = ['sniffer.py']
    cmd += [
        '-c', channel, '-u', interface_port, '--crc', '--rssi', '-b',
        interface_baudrate, '-o',
        str(fifo), '--is-fifo', '--use-host-timestamp'
    ]
    if tap:
        cmd.append('--tap')

    subprocess.Popen(cmd).wait()


def extcap_close_fifo(fifo):
    """"Close extcap fifo"""
    # This is apparently needed to workaround an issue on Windows/macOS
    # where the message cannot be read. (really?)
    fh = open(fifo, 'wb', 0)
    fh.close()


if __name__ == '__main__':

    # Capture options
    parser = argparse.ArgumentParser(
        description='OpenThread Sniffer extcap plugin')

    # Extcap Arguments
    parser.add_argument('--extcap-interfaces',
                        help='Provide a list of interfaces to capture from',
                        action='store_true')
    parser.add_argument('--extcap-interface',
                        help='Provide the interface to capture from')
    parser.add_argument('--extcap-dlts',
                        help='Provide a list of dlts for the given interface',
                        action='store_true')
    parser.add_argument(
        '--extcap-config',
        help='Provide a list of configurations for the given interface',
        action='store_true')
    parser.add_argument('--extcap-reload-option',
                        help='Reload elements for the given option')
    parser.add_argument('--capture',
                        help='Start the capture routine',
                        action='store_true')
    parser.add_argument(
        '--fifo',
        help='Use together with capture to provide the fifo to dump data to')
    parser.add_argument(
        '--extcap-capture-filter',
        help='Used together with capture to provide a capture filter')
    parser.add_argument('--extcap-control-in',
                        help='Used to get control messages from toolbar')
    parser.add_argument('--extcap-control-out',
                        help='Used to send control messages to toolbar')
    parser.add_argument('--extcap-version', help='Wireshark Version')

    # Interface Arguments
    parser.add_argument('--channel',
                        help='IEEE 802.15.4 capture channel [11-26]')
    parser.add_argument(
        '--tap',
        help='IEEE 802.15.4 TAP (only for Wireshark3.0 and later)',
        action='store_true')

    try:
        args, unknown = parser.parse_known_args()
    except argparse.ArgumentError as e:
        parser.exit('ERROR_ARG: %s' % str(e))

    extcap_version = ''
    version_path = os.path.join(tempfile.gettempdir(), 'extcap_ot_version')
    if args.extcap_version:
        extcap_version = args.extcap_version
        with open(version_path, mode='w') as f:
            f.write(extcap_version)
    else:
        try:
            with open(version_path, mode='r') as f:
                extcap_version = f.read()
        except FileNotFoundError:
            pass

    if len(unknown) > 0:
        parser.exit('Sniffer %d unknown arguments given: %s' %
                    (len(unknown), unknown))

    if len(sys.argv) == 0:
        parser.print_help()
        parser.exit('No arguments given!')

    if not args.extcap_interfaces and args.extcap_interface is None:
        parser.exit(
            'An interface must be provided or the selection must be displayed')

    if args.extcap_interfaces:
        extcap_interfaces()
        sys.exit(0)

    if args.extcap_config:
        extcap_config(args.extcap_interface, '', extcap_version)
    elif args.extcap_dlts:
        extcap_dlts(args.extcap_interface)
    elif args.capture:
        if args.fifo is None:
            parser.exit('The fifo must be provided to capture')
        try:
            extcap_capture(args.extcap_interface, args.fifo,
                           args.extcap_control_in, args.extcap_control_out,
                           args.channel, args.tap)
        except KeyboardInterrupt:
            pass
        except Exception as e:
            logging.exception(e)
            parser.exit('ERROR_INTERNAL')
    else:
        parser.print_help()
        parser.exit('ERROR_USAGE')
